import secrets

from hipercow.driver import load_driver_optional
from hipercow.environment import environment_check
from hipercow.resources import TaskResources
from hipercow.root import OptionalRoot, Root, open_root
from hipercow.task import TaskData, TaskStatus, set_task_status
from hipercow.util import relative_workdir


def task_create_shell(
    cmd: list[str],
    *,
    environment: str | None = None,
    envvars: dict[str, str] | None = None,
    resources: TaskResources | None = None,
    driver: str | None = None,
    root: OptionalRoot = None,
) -> str:
    """Create a shell command task.

    This is the first type of task that we support, and more types
    will likely follow.  A shell command will evaluate an arbitrary
    command on the cluster - it does not even need to be written in
    Python! However, if you are using the `pip` environment engine
    then it will need to be `pip`-installable.

    The interface here is somewhat subject to change, but we think the
    basics here are reasonable.

    Args:
        cmd: The command to execute, as a list of strings

        environment: The name of the environment to evaluate the
            command in.  The default (`None`) will select `default` if
            available, falling back on `empty`.

        envvars: A dictionary of environment variables to set before
            the task runs.  Do not set `PATH` in here, it will not
            currently have an effect.

        resources: Optional resources required by your task.

        driver: The driver to launch the task with.  Generally this is
            not needed as we expect most people to have a single
            driver set.

        root: The root, or if not given search from the current directory.

    Returns:
        The newly-created task identifier, a 32-character hex string.
    """
    root = open_root(root)
    if not cmd:
        msg = "'cmd' cannot be empty"
        raise Exception(msg)
    data = {"cmd": cmd}
    task_id = _task_create(
        root=root,
        method="shell",
        environment=environment,
        driver=driver,
        data=data,
        resources=resources,
        envvars=envvars or {},
    )
    return task_id


def _task_create(
    *,
    root: Root,
    method: str,
    environment: str | None,
    driver: str | None,
    data: dict,
    resources: TaskResources | None,
    envvars: dict[str, str],
) -> str:
    path = relative_workdir(root.path)
    task_id = _new_task_id()
    environment = environment_check(environment, root)
    dr = load_driver_optional(driver, root)
    if resources:
        if not dr:
            msg = "Can't specify resources, as driver is not given"
            raise Exception(msg)
        resources = dr.resources().validate_resources(resources)
    task_data = TaskData(
        task_id=task_id,
        method=method,
        data=data,
        path=str(path),
        environment=environment,
        resources=resources,
        envvars=envvars,
    )
    task_data.write(root)
    with root.path_recent().open("a") as f:
        f.write(f"{task_id}\n")
    if dr:
        dr.submit(task_id, resources, root)
        set_task_status(task_id, TaskStatus.SUBMITTED, dr.name, root)
    return task_id


def _new_task_id() -> str:
    return secrets.token_hex(16)



"""Interact with the hipercow root."""

import platform
from pathlib import Path
from typing import TypeAlias

from hipercow import ui
from hipercow.util import find_file_descend


def init(path: str | Path) -> None:
    """Initialise a hipercow root.

    Sets up a new hipercow root at `path`, creating the directory
    `<path>/hipercow/` which will contain all of hipercow's files.  It
    is safe to re-initialise an already-created hipercow root (in
    which case nothing happens) and safe to initialise a *Python*
    hipercow root at the same location as an R one, though at the
    moment they do not interact.

    Args:
        path: The path to the project root

    Returns:
        Nothing, called for side effects only.

    """
    path = Path(path)
    dest = path / "hipercow" / "py"

    if dest.is_dir():
        ui.alert_warning(f"hipercow already initialised at '{path.resolve()}'")
        return

    root = find_file_descend("hipercow", path)
    if root is not None:
        ui.alert_warning(
            f"hipercow already initialised at '{root}' (found from '{path}')"
        )

    if dest.exists() and not dest.is_dir():
        msg = (
            "Unexpected file 'hipercow/py' (rather than directory)"
            f"found at {path}"
        )
        raise Exception(msg)

    dest.mkdir(parents=True)
    _add_gitignore(dest.parent)
    ui.alert_success(f"Initialised hipercow at '{path.resolve()}'")


class Root:
    def __init__(self, path: str | Path) -> None:
        self.path = Path(path)
        if not self.path_base().parent.is_dir():
            msg = f"Failed to open 'hipercow' root at {path}"
            raise Exception(msg)
        if not self.path_base().is_dir():
            msg = f"Failed to open non-python 'hipercow' root at {path}"
            raise Exception(msg)

    def path_base(self, *, relative: bool = False) -> Path:
        return (Path() if relative else self.path) / "hipercow" / "py"

    def path_task(self, task_id: str | None, *, relative: bool = False) -> Path:
        base = self.path_base(relative=relative)
        if task_id is None:
            return base / "tasks"
        return base / "tasks" / task_id[:2] / task_id[2:]

    def path_task_times(self, task_id: str) -> Path:
        return self.path_task(task_id) / "times"

    def path_task_data(self, task_id: str) -> Path:
        return self.path_task(task_id) / "data"

    def path_task_result(self, task_id: str) -> Path:
        return self.path_task(task_id) / "result"

    def path_task_log(self, task_id: str) -> Path:
        return self.path_task(task_id) / "log"

    def path_recent(self) -> Path:
        return self.path_base() / "recent"

    def path_configuration(self, name: str | None) -> Path:
        hostname = platform.node()
        return self.path_base() / "config" / hostname / (name or ".")

    def path_environment(
        self, name: str | None, *, relative: bool = False
    ) -> Path:
        base = self.path_base(relative=relative)
        return base / "env" / (name or ".")

    def path_environment_config(self, name: str) -> Path:
        return self.path_environment(name) / "config"

    def path_environment_contents(
        self, name: str, *, relative: bool = False
    ) -> Path:
        return self.path_environment(name, relative=relative) / "contents"

    def path_provision(
        self, name: str, id: str, *, relative: bool = False
    ) -> Path:
        return self.path_environment(name, relative=relative) / "provision" / id

    def path_provision_data(
        self, name: str, id: str, *, relative: bool = False
    ) -> Path:
        return self.path_provision(name, id, relative=relative) / "data"

    def path_provision_result(
        self, name: str, id: str, *, relative: bool = False
    ) -> Path:
        return self.path_provision(name, id, relative=relative) / "result"

    def path_provision_log(
        self, name: str, id: str, *, relative: bool = False
    ) -> Path:
        return self.path_provision(name, id, relative=relative) / "log"

    def path_repl_history(self) -> Path:
        return self.path_base() / "repl_history"

    def path_bundle(self, name: str | None) -> Path:
        return self.path_base() / "bundles" / (name or ".")


OptionalRoot: TypeAlias = None | str | Path | Root
"""Optional root type, for user-facing functions.

Represents different inputs to the user-facing functions in hipercow,
recognising that most of the time the working directory will be within
a hipercow environment.  The possible types represent different
possibilities:

* `None`: search for the root from the current directory (like `git` does)
* `str`: the name of a path to search from
* `Path`: a `pathlib.Path` object representing the path to search from
* `Root`: a root previously opened with [`open_root`][hipercow.root.open_root]

"""


def open_root(path: OptionalRoot = None) -> Root:
    """Open a hipercow root.

    Locate and validate a hipercow root, converting an `OptionalRoot` type
    into a real `Root` object.  This function is used in most user-facing
    functions in hipercow, but you can call it yourself to validate the
    root early.

    Args:
        path: A path to the root to open, a `Root`, or `None`.

    Returns:
        The opened `Root` object.

    """
    if isinstance(path, Root):
        return path
    root = find_file_descend("hipercow", path or Path.cwd())
    if not root:
        msg = f"Failed to find 'hipercow' from {path}"
        raise Exception(msg)
    return Root(root)


def _add_gitignore(path: Path):
    dest = path / ".gitignore"
    if not dest.exists():
        with dest.open("w") as f:
            f.write("*\n")



"""Specify and interact with resources."""

import math
from dataclasses import dataclass


@dataclass(kw_only=True)
class TaskResources:
    """Resources required for a task.

    We don't support everything that the R version does yet; in
    particular we've not set up `hold_until`, `priority` or
    `requested_nodes`, as these are not widely used.

    Attributes:
        queue: The queue to run on.  If not given (or `None`), we use
            the default queue for your cluster.  Alternatively, you
            can provide `.default` for the default queue or `.test`
            for the test queue.

        cores: The number of cores to request.  Adding more cores does
            not necessarily make your task any faster; your task must
            have some mechanism to exploit this parallelism (e.g.,
            using the
            [`multiprocessing`](https://docs.python.org/3/library/multiprocessing.html)
            package).  Specify `math.inf` if you want to request all
            the cores on a single node.

        exclusive: Request exclusive access to a node.

        max_runtime: The maximum run time (wall clock), in seconds.

        memory_per_node: Specify that your task can only run on a node
            with at least this much memory, in GB (e.g, 128 is 128GB).

        memory_per_task: An estimate of how much memory your task
            requires (across all cores), in GB.  If you provide this,
            the scheduler can attempt to arrange tasks such that they
            will all fit in the available RAM.

    """

    queue: str | None = None
    cores: int | float = 1
    exclusive: bool = False
    max_runtime: int | None = None
    memory_per_node: int | None = None
    memory_per_task: int | None = None

    def __post_init__(self):
        _require_positive(self.cores, "cores")
        if self.cores != math.inf and not isinstance(self.cores, int):
            msg = "'cores' must be an integer (or inf)"
            raise ValueError(msg)
        _require_positive(self.max_runtime, "max_runtime")
        _require_positive(self.memory_per_node, "memory_per_node")
        _require_positive(self.memory_per_task, "memory_per_task")


@dataclass
class Queues:
    """Queues available on the cluster.

    Attributes:
        valid: The set of valid queue names. Being a set, the order does
            not imply anything.

        default: The default queue, used if none is explicitly given

        build: The queue used to run build jobs

        test: The queue used to run test jobs
    """

    valid: set[str]
    default: str
    build: str
    test: str

    @staticmethod
    def simple(name: str) -> "Queues":
        """Create a `Queues` object with only one valid queue.

        This situation is common enough that we provide a small wrapper.

        Args:
            name: The only supported queue.  This will become the set
                of valid queues, the default queue, the build queue and
                the test queue.
        """
        return Queues({name}, name, name, name)

    def __post_init__(self):
        self._check_queue(self.default, "Default queue")
        self._check_queue(self.build, "Build queue")
        self._check_queue(self.test, "Test queue")

    def _check_queue(self, name: str, description: str = "Queue"):
        if name not in self.valid:
            msg = f"{description} '{name}' is not in valid queue list"
            raise ValueError(msg)

    def validate_queue(
        self, name: str | None, description: str = "Queue"
    ) -> str:
        if name is None:
            return self.default
        if name.startswith("."):
            if name[1:] == "default":
                return self.default
            elif name[1:] == "test":
                return self.test
            else:
                msg = f"Invalid special queue '{name}'"
                raise ValueError(msg)
        self._check_queue(name, description)
        return name


@dataclass
class ClusterResources:
    """Resources available on a cluster.

    This will be returned by a cluster driver and will be used to
    validate resources.

    Attributes:
       queues: Valid queues.
       max_memory: The maximum ram across all nodes in the cluster.
       max_cores: The maximum cores across all nodes in the cluster.

    """

    queues: Queues
    max_cores: int
    max_memory: int

    def validate_resources(self, resources: TaskResources) -> TaskResources:
        """Check resources are valid on this cluster.

        Takes a set of resources and checks that the requested queue,
        number of cores and ram are valid for the cluster.  If the
        queue is not provided, then we will take the default.

        Args:
            resources: Resources to validate.

        """
        resources.queue = self.queues.validate_queue(resources.queue)
        if resources.cores > self.max_cores:
            msg = f"{resources.cores} is too many cores for this cluster"
            raise ValueError(msg)
        _check_ram(resources.memory_per_node, self.max_memory, "node")
        _check_ram(resources.memory_per_task, self.max_memory, "task")

        return resources


def _require_positive(x: float | int, name: str) -> None:
    if x is not None and x < 0:
        msg = f"'{name}' must be positive"
        raise ValueError(msg)


def _check_ram(requested: int | None, available: int, type: str):
    if requested is not None and requested > available:
        msg = f"{requested}Gb per {type} is too large for this cluster"
        raise ValueError(msg)



from pathlib import Path

from taskwait import Task, taskwait

from hipercow import ui
from hipercow.dide.auth import fetch_credentials
from hipercow.dide.batch import write_batch_provision, write_batch_task_run
from hipercow.dide.configuration import DideConfiguration
from hipercow.dide.mounts import detect_mounts
from hipercow.dide.web import DideWebClient
from hipercow.driver import HipercowDriver
from hipercow.resources import ClusterResources, Queues, TaskResources
from hipercow.root import Root


class DideWindowsDriver(HipercowDriver):
    name = "dide-windows"
    config: DideConfiguration

    def __init__(self, root: Root, **kwargs):
        mounts = detect_mounts()
        self.config = DideConfiguration(root, mounts=mounts, **kwargs)

    def show_configuration(self) -> None:
        path_map = self.config.path_map
        ui.li("[bold]Path mapping[/bold]")
        ui.li(f"drive '{path_map.remote}'", indent=2, symbol="-")
        ui.li(
            f"share '\\\\{path_map.mount.host}\\{path_map.mount.remote}'",
            indent=2,
            symbol="-",
        )
        ui.li(f"[bold]Python version[/bold]: {self.config.python_version}")

    def submit(
        self, task_id: str, resources: TaskResources | None, root: Root
    ) -> None:
        cl = _web_client()
        unc = write_batch_task_run(task_id, self.config, root)
        if not resources:
            resources = self.resources().validate_resources(TaskResources())
        dide_id = cl.submit(unc, task_id, resources=resources)
        with self._path_dide_id(task_id, root).open("w") as f:
            f.write(dide_id)

    def provision(self, name: str, id: str, root: Root) -> None:
        _dide_provision(name, id, self.config, root)

    def resources(self) -> ClusterResources:
        # We should get this from the cluster itself but with caching
        # not yet configured this seems unwise as we'll hit the
        # cluster an additional time for every job submission rather
        # than just once a session.
        queues = Queues(
            {"AllNodes", "BuildQueue", "Testing"},
            default="AllNodes",
            test="Testing",
            build="BuildQueue",
        )
        return ClusterResources(queues=queues, max_cores=32, max_memory=512)

    def task_log(
        self, task_id: str, *, outer: bool = False, root: Root
    ) -> str | None:
        if outer:
            with self._path_dide_id(task_id, root).open() as f:
                dide_id = f.read().strip()
            cl = _web_client()
            return cl.log(dide_id.strip())
        return super().task_log(task_id, outer=False, root=root)

    def _path_dide_id(self, task_id: str, root: Root) -> Path:
        return root.path_task(task_id) / "dide_id"


class ProvisionWaitWrapper(Task):
    def __init__(
        self,
        root: Root,
        name: str,
        provision_id: str,
        client: DideWebClient,
        dide_id: str,
    ):
        self.root = root
        self.name = name
        self.provision_id = provision_id
        self.client = client
        self.dide_id = dide_id
        self.status_waiting = {"created", "submitted"}
        self.status_running = {"running"}

    def status(self) -> str:
        return str(self.client.status_job(self.dide_id))

    def log(self) -> list[str] | None:
        path = self.root.path_provision_log(self.name, self.provision_id)
        if not path.exists():
            return None
        with path.open() as f:
            return f.read().splitlines()

    def has_log(self) -> bool:
        return True


def _web_client() -> DideWebClient:
    credentials = fetch_credentials()
    cl = DideWebClient(credentials)
    cl.login()
    return cl


def _dide_provision(name: str, id: str, config: DideConfiguration, root: Root):
    cl = _web_client()
    unc = write_batch_provision(name, id, config, root)
    resources = TaskResources(queue="BuildQueue")
    dide_id = cl.submit(unc, f"{name}/{id}", resources=resources)
    task = ProvisionWaitWrapper(root, name, id, cl, dide_id)
    res = taskwait(task)
    dt = round(res.end - res.start, 2)
    if res.status == "failure":
        path_log = root.path_provision_log(name, id, relative=True)
        ui.alert_danger(f"Provisioning failed after {dt}s!")
        ui.blank_line()
        ui.text("Logs, if produced, may be visible above")
        ui.text("A copy of all logs is available at:")
        ui.text(f"    {path_log}")
        ui.blank_line()
        dide_log = cl.log(dide_id)
        ui.logs("Logs from the cluster", dide_log)
        msg = "Provisioning failed"
        raise Exception(msg)
    else:
        ui.alert_success(f"Provisioning completed in {dt}s")



import pytest

from hipercow import root
from hipercow import task_create as tc
from hipercow.task import TaskStatus, task_log, task_status
from hipercow.task_eval import task_eval
from hipercow.util import transient_working_directory


def test_can_set_task_status(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)
    assert task_status(tid, r) == TaskStatus.CREATED
    task_eval(tid, capture=False, root=r)
    assert task_status(tid, r) == TaskStatus.SUCCESS


def test_cant_run_complete_task(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)
    task_eval(tid, capture=False, root=r)
    msg = f"Can't run '{tid}', which has status 'success'"
    with pytest.raises(Exception, match=msg):
        task_eval(tid, capture=False, root=r)


def test_can_capture_output_to_auto_file(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)
    task_eval(tid, capture=True, root=r)

    path = r.path_task_log(tid)
    with path.open("r") as f:
        assert f.read().strip() == "hello world"

    assert task_log(tid, root=r) == "hello world\n"


def test_return_information_about_failure_to_find_path(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["adsfasdfasdfa", "arg"], root=r)
    task_eval(tid, capture=True, root=r)

    path = r.path_task_log(tid)
    assert path.exists()
    assert task_status(tid, r) == TaskStatus.FAILURE



from hipercow import root
from hipercow import task_create as tc
from hipercow.dide import batch
from hipercow.dide.configuration import DideConfiguration
from hipercow.dide.mounts import Mount
from hipercow.util import transient_working_directory


def test_can_create_batch_data(tmp_path):
    path = tmp_path / "my/project"
    root.init(path)
    r = root.open_root(path)
    m = Mount("wpia-hn", "didehomes/bob", tmp_path)
    config = DideConfiguration(
        r, mounts=[m], python_version=None, check_credentials=False
    )

    res = batch._template_data_task_run("abcde", config)
    assert res["task_id"] == "abcde"
    assert res["task_id_1"] == "ab"
    assert res["task_id_2"] == "cde"
    assert res["hipercow_root_drive"] == "V:"
    assert res["hipercow_root_path"] == "\\my\\project"
    assert (
        res["network_shares_create"] == r"net use V: \\wpia-hn\didehomes\bob /y"
    )
    assert res["network_shares_delete"] == "net use V: /delete /y"


def test_can_write_batch(tmp_path):
    path = tmp_path / "my/project"
    root.init(path)
    r = root.open_root(path)
    m = Mount("wpia-hn", "didehomes/bob", tmp_path)
    config = DideConfiguration(
        r, mounts=[m], python_version=None, check_credentials=False
    )

    with transient_working_directory(path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)

    unc = batch.write_batch_task_run(tid, config, r)
    path_rel = f"hipercow\\py\\tasks\\{tid[:2]}\\{tid[2:]}\\task_run.bat"
    assert unc == f"\\\\wpia-hn\\didehomes\\bob\\my\\project\\{path_rel}"
    assert (r.path / path_rel.replace("\\", "/")).exists()


def test_can_create_provision_data(tmp_path):
    path = tmp_path / "my/project"
    root.init(path)
    r = root.open_root(path)
    m = Mount("wpia-hn", "didehomes/bob", tmp_path)
    config = DideConfiguration(
        r, mounts=[m], python_version=None, check_credentials=False
    )

    res = batch._template_data_provision("env", "abcde", config)
    assert res["environment_name"] == "env"
    assert res["provision_id"] == "abcde"
    assert res["hipercow_root_drive"] == "V:"
    assert res["hipercow_root_path"] == "\\my\\project"
    assert (
        res["network_shares_create"] == r"net use V: \\wpia-hn\didehomes\bob /y"
    )
    assert res["network_shares_delete"] == "net use V: /delete /y"


def test_can_write_provision_batch(tmp_path):
    path = tmp_path / "my/project"
    root.init(path)
    r = root.open_root(path)
    m = Mount("wpia-hn", "didehomes/bob", tmp_path)
    config = DideConfiguration(
        r, mounts=[m], python_version=None, check_credentials=False
    )

    unc = batch.write_batch_provision("myenv", "abcdef", config, r)
    path_rel = "hipercow\\py\\env\\myenv\\provision\\abcdef\\run.bat"
    assert unc == f"\\\\wpia-hn\\didehomes\\bob\\my\\project\\{path_rel}"
    assert (r.path / path_rel.replace("\\", "/")).exists()



import csv
import os
import platform
import re
import subprocess
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from dataclasses import dataclass
from itertools import product
from pathlib import Path
from typing import Any


def find_file_descend(filename: str, path: str | Path) -> Path | None:
    path = Path(path)
    root = Path(path.anchor)

    while path != root:
        attempt = path / filename
        if attempt.exists():
            return attempt.parent
        path = path.parent

    return None


def relative_workdir(path: str | Path, base: None | str | Path = None) -> Path:
    return Path(path).relative_to(Path(base) if base else Path.cwd())


@contextmanager
def transient_working_directory(path):
    origin = os.getcwd()
    try:
        if path is not None:
            os.chdir(path)
        yield
    finally:
        if path is not None:
            os.chdir(origin)


@contextmanager
def transient_envvars(env: dict[str, str | None]) -> Iterator[None]:
    def _set_envvars(env):
        for k, v in env.items():
            if v is None:
                os.environ.pop(k, None)
            else:
                os.environ[k] = v

    prev = {k: os.environ.get(k) for k in env.keys()}
    try:
        _set_envvars(env)
        yield
    finally:
        _set_envvars(prev)


def file_create(path: str | Path) -> None:
    Path(path).open("a").close()


def subprocess_run(
    cmd,
    *,
    filename: Path | None = None,
    check=False,
    env: dict | None = None,
    **kwargs,
) -> subprocess.CompletedProcess:
    env = os.environ | (env or {})
    try:
        if filename is None:
            return subprocess.run(cmd, **kwargs, check=check, env=env)
        else:
            with filename.open("ab") as f:
                return subprocess.run(
                    cmd,
                    check=check,
                    env=env,
                    stderr=subprocess.STDOUT,
                    stdout=f,
                    **kwargs,
                )
    except FileNotFoundError as err:
        if check:
            raise err
        elif filename is None:
            print(err)
        else:
            with filename.open("a") as f:
                print(err, file=f)
        return subprocess.CompletedProcess(cmd, -1)


def check_python_version(
    version: str | None, valid: list[str] | None = None
) -> str:
    if valid is None:
        valid = ["3.10", "3.11", "3.12", "3.13"]
    if not version:
        v = ".".join(platform.python_version_tuple()[:2])
    else:
        m = re.match(r"^([0-9]+)\.([0-9]+)(\.[0-9]+)?$", version)
        if not m:
            msg = f"'{version}' does not parse as a valid version string"
            raise Exception(msg)
        v = ".".join(m.groups()[:2])
    if v not in valid:
        msg = f"Version '{version}' is not supported"
        raise Exception(msg)
    return v


def truthy_envvar(name: str) -> bool:
    value = os.environ.get(name)
    return value is not None and (value.lower() in {"1", "true"})


def read_file_if_exists(path: Path) -> str | None:
    if not path.exists():
        return None
    with path.open() as f:
        return f.read()


def loop_while(fn: Callable[[], bool]) -> None:
    while True:
        if not fn():
            break


@dataclass
class Result:
    exception: Exception | None = None

    def __bool__(self) -> bool:
        return self.exception is None

    @staticmethod
    def ok() -> "Result":
        return Result()

    @staticmethod
    def err(exception: Exception) -> "Result":
        return Result(exception)


def expand_grid(data: dict) -> list[dict]:
    return [
        dict(zip(data.keys(), el, strict=False))
        for el in product(*data.values())
    ]


# Probably some more work here to get the name to str here?
def read_csv_to_dict(filename: str | Path) -> list[dict[str, Any]]:
    with Path(filename).open(newline="") as f:
        return list(csv.DictReader(f))


# We could make this more generic, but this changes syntax at 3.12
# https://mypy.readthedocs.io/en/stable/generics.html#generic-functions
def tabulate(x: list[str]) -> dict[str, int]:
    ret: dict[str, int] = {}
    for el in x:
        try:
            ret[el] += 1
        except KeyError:
            ret[el] = 1
    return ret



from hipercow import ui
from hipercow.driver import HipercowDriver
from hipercow.provision import provision_run
from hipercow.resources import ClusterResources, Queues, TaskResources
from hipercow.root import Root
from hipercow.util import check_python_version


class ExampleDriver(HipercowDriver):
    name = "example"

    def __init__(self, root: Root, **kwargs):  # noqa: ARG002
        version = kwargs.get("python_version")
        if isinstance(version, str):
            requested = check_python_version(version)
            local = check_python_version(None)
            if local != requested:
                msg = (
                    f"Requested python version {version}"
                    f"is not the same as the local version {local}"
                )
                raise Exception(msg)

    def show_configuration(self) -> None:
        ui.text("[dim](no configuration)[/dim]")

    def submit(
        self,
        task_id: str,
        resources: TaskResources | None,  # noqa: ARG002
        root: Root,  # noqa: ARG002
    ) -> None:
        ui.alert_info(f"submitting '{task_id}'")

    def provision(self, name: str, id: str, root: Root) -> None:
        provision_run(name, id, root)

    def resources(self) -> ClusterResources:
        return ClusterResources(
            queues=Queues.simple("default"),
            max_cores=1,
            max_memory=32,
        )



import pickle
import shutil
from dataclasses import dataclass

from hipercow import ui
from hipercow.environment_engines import (
    Empty,
    EnvironmentEngine,
    Pip,
)
from hipercow.root import OptionalRoot, Root, open_root


@dataclass
class EnvironmentConfiguration:
    engine: str

    # As with elsewhere, we will need to avoid actually serialising
    # the instance itself and only the configuration. Ignore this for
    # now, even though this will create versioning headaches for us.
    def write(self, name: str, root: Root):
        path = root.path_environment_config(name)
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def read(name: str, root: Root) -> "EnvironmentConfiguration":
        with root.path_environment_config(name).open("rb") as f:
            return pickle.load(f)


# Called 'new' and not 'create' to make it clear that this does not
# actually create the environment, just the definition of that
# environment.
def environment_new(name: str, engine: str, root: OptionalRoot = None) -> None:
    """Create a new environment.

    Creating an environment selects a name and declares the engine for
    the environment.  After doing this, you will certainly want to
    provision the environment using `provision()`.

    Args:
        name: The name for the environment.  The name `default` is a
            good choice if you only want a single environment, as this
            is the environment used by default.  You cannot use
            `empty` as that is a special empty environment.

        engine: The environment engine to use.  The options here are
            `pip` and `empty`.  Soon we will support `conda` too.

        root: The root, or if not given search from the current directory.

    Returns:
        Nothing, called for side effects.

    """
    root = open_root(root)
    path = root.path_environment(name)

    # We might make this friendlier later
    if name == "empty" or path.exists():
        msg = f"Environment '{name}' already exists"
        raise Exception(msg)

    if engine not in {"pip", "empty"}:
        msg = "Only the 'pip' and 'empty' engines are supported"
        raise Exception(msg)

    ui.alert_info(f"Creating environment '{name}' using '{engine}'")
    EnvironmentConfiguration(engine).write(name, root)


def environment_list(root: OptionalRoot = None) -> list[str]:
    """List known environments.

    Args:
        root: The root, or if not given search from the current directory.

    Returns:
        A sorted list of environment names.  The name `empty` will
            always be present.

    """
    root = open_root(root)
    special = ["empty"]
    path = root.path_environment(None)
    found = [x.name for x in path.glob("*")]
    return sorted(special + found)


def environment_delete(name: str, root: OptionalRoot = None) -> None:
    """Delete an environment.

    Args:
        name: The name of the environment to delete.
        root: The root, or if not given search from the current directory.

    Returns:
        Nothing, called for side effects only.
    """
    root = open_root(root)
    if name == "empty":
        msg = "Can't delete the empty environment"
        raise Exception(msg)
    if not environment_exists(name, root):
        if name == "default":
            reason = "it is empty"
        else:
            reason = "it does not exist"
        msg = f"Can't delete environment '{name}', as {reason}"
        raise Exception(msg)
    ui.alert_info(
        f"Attempting to delete environment '{name}'; this might fail if "
        "files are in use on a network share, in which case you should "
        "try again later",
    )
    shutil.rmtree(str(root.path_environment(name)))
    ui.alert_success("Done!")


def environment_check(name: str | None, root: OptionalRoot = None) -> str:
    """Validate an environment name for this root.

    This function can be used to ensure that `name` is a reasonable
    environment name to use in your root.  It returns the resolved
    name (selecting between `empty` and `default` if `name` is
    `None`), and errors if the requested environment is not found.

    Args:
        name: The name of the environment to use, or `None` to select
            the appropriate default.
        root: The root, or if not given search from the current directory.

    Returns:
        The resolved environment name.

    """
    root = open_root(root)
    if name is None:
        return "default" if environment_exists("default", root) else "empty"
    if name == "empty" or environment_exists(name, root):
        return name
    msg = f"No such environment '{name}'"
    raise Exception(msg)


def environment_exists(name: str, root: OptionalRoot = None) -> bool:
    """Check if an environment exists.

    Note that this function will return `False` for `empty`, even
    though `empty` is always a valid choice.  We might change this in
    future.

    Args:
        name: The name of the environment to check.
        root: The root, or if not given search from the current directory.

    Returns:
        `True` if the environment exists, otherwise `False`.

    """
    root = open_root(root)
    return root.path_environment(name).exists()


# TODO: move this somewhere less user-facing
def environment_engine(name: str, root: Root) -> EnvironmentEngine:
    use_empty_environment = name == "empty" or (
        name == "default" and not environment_exists(name, root)
    )
    if use_empty_environment:
        cfg = EnvironmentConfiguration("empty")
    else:
        cfg = EnvironmentConfiguration.read(name, root)
    if cfg.engine == "pip":
        return Pip(root, name)
    elif cfg.engine == "empty":
        return Empty(root, name)
    raise NotImplementedError()  # pragma no cover



import subprocess
from pathlib import Path

from hipercow.environment_engines.base import EnvironmentEngine, Platform
from hipercow.root import Root
from hipercow.util import subprocess_run


class Empty(EnvironmentEngine):
    """The empty environment, into which nothing may be installed."""

    def __init__(self, root: Root, name: str, platform: Platform | None = None):
        super().__init__(root, name, platform)

    def exists(self) -> bool:
        """The empty environment always exists.

        Returns:
            `True`

        """
        return True

    def path(self) -> Path:
        """The path to the empty environment, which never exists.

        Returns:
           Never returns, but throws if called.
        """
        msg = "The empty environment has no path"
        raise Exception(msg)

    # These "unused argument" errors from ruff are probably a bug?
    def create(self, **kwargs) -> None:  # noqa: ARG002
        """Create the empty environment, which already exists.

        Returns:
            Never returns, but throws if called.
        """
        msg = "Can't create the empty environment!"
        raise Exception(msg)

    def check_args(self, cmd: list[str] | None) -> list[str]:
        """Check arguments to the empty environment, which must be empty.

        Args:
            cmd: A list or `None`, if the list is not empty an error is thrown.

        Returns:
            The empty list.
        """
        if not cmd:
            return []
        else:
            msg = "No arguments are allowed to the empty environment"
            raise Exception(msg)

    def provision(self, cmd: list[str], **kwargs) -> None:  # noqa: ARG002
        """Install packages into the empty environment, which is not allowed.

        Returns:
            Never returns, but throws if called.
        """
        msg = "Can't provision the empty environment!"
        raise Exception(msg)

    def run(
        self,
        cmd: list[str],
        *,
        env: dict[str, str] | None = None,
        **kwargs,
    ) -> subprocess.CompletedProcess:
        return subprocess_run(cmd, env=env, **kwargs)



import shutil

import pytest

from hipercow import root, util


def test_create_root(tmp_path):
    path = tmp_path / "ex"
    root.init(path)
    assert path.exists()
    assert path.is_dir()
    assert (path / "hipercow" / "py").is_dir()
    r = root.open_root(path)
    assert isinstance(r, root.Root)
    assert r.path == path


def test_notify_if_root_exists(tmp_path, capsys):
    path = tmp_path
    root.init(path)
    capsys.readouterr()
    root.init(path)
    captured = capsys.readouterr()
    assert "hipercow already initialised at" in captured.out


def test_notify_if_root_exists_below_requested(tmp_path, capsys):
    path = tmp_path / "a" / "b"
    root.init(tmp_path)
    capsys.readouterr()
    root.init(path)
    captured = capsys.readouterr()
    assert "hipercow already initialised at" in captured.out


def test_error_if_root_invalid(tmp_path):
    (tmp_path / "hipercow").mkdir()
    util.file_create(tmp_path / "hipercow" / "py")
    with pytest.raises(Exception, match="Unexpected file 'hipercow/py'"):
        root.init(tmp_path)


def test_error_if_root_does_not_exist(tmp_path):
    with pytest.raises(Exception, match="Failed to open 'hipercow' root"):
        root.Root(tmp_path)


def test_error_if_non_python_hipercow_root_found(tmp_path):
    root.init(tmp_path)
    shutil.rmtree(tmp_path / "hipercow" / "py")
    pat = "Failed to open non-python 'hipercow' root at"
    with pytest.raises(Exception, match=pat):
        root.Root(tmp_path)


def test_find_root_by_descending(tmp_path):
    path = tmp_path / "a" / "b"
    root.init(tmp_path)
    r = root.open_root(path)
    assert r.path == tmp_path


def test_error_if_no_root_found_by_descending(tmp_path):
    path = tmp_path / "a" / "b"
    with pytest.raises(Exception, match="Failed to find 'hipercow' from"):
        root.open_root(path)


def test_create_gitignore_in_root(tmp_path):
    path = tmp_path / "ex"
    gi = path / "hipercow" / ".gitignore"
    root.init(path)
    assert gi.exists()
    with gi.open() as f:
        assert f.read() == "*\n"


def test_dont_overwrite_existing_gitignore(tmp_path):
    path = tmp_path / "ex"
    gi = path / "hipercow" / ".gitignore"
    (path / "hipercow").mkdir(parents=True)
    with gi.open("w") as f:
        f.write("hello!\n")
    root.init(path)
    assert gi.exists()
    with gi.open() as f:
        assert f.read() == "hello!\n"



def pytest_addoption(parser):
    parser.addoption(
        "--longrun",
        action="store_true",
        dest="longrun",
        default=False,
        help="enable longrundecorated tests",
    )



from unittest import mock

from hipercow import ui


def test_can_build_ui_styles(mocker):
    mocker.patch("hipercow.ui.console")

    ui.alert_danger("oh no")
    assert ui.console.print.call_count == 1
    assert ui.console.print.mock_calls[0] == mock.call(
        "[bold red]:heavy_multiplication_x:[/bold red] oh no"
    )

    ui.alert_success("yaaassss.", indent=2)
    assert ui.console.print.call_count == 2
    assert ui.console.print.mock_calls[1] == mock.call(
        "  [bold green]:heavy_check_mark:[/bold green] yaaassss."
    )

    ui.alert_warning("check it before you wreck it")
    assert ui.console.print.call_count == 3
    assert ui.console.print.mock_calls[2] == mock.call(
        "[bold orange]![/bold orange] check it before you wreck it"
    )

    ui.alert_info("here it is")
    assert ui.console.print.call_count == 4
    assert ui.console.print.mock_calls[3] == mock.call(
        "[bold cyan]i[/bold cyan] here it is"
    )

    ui.alert_see_also("this doc")
    assert ui.console.print.call_count == 5
    assert ui.console.print.mock_calls[4] == mock.call(
        ":books: For more information, see this doc"
    )

    ui.alert_arrow("hello")
    assert ui.console.print.call_count == 6
    assert ui.console.print.mock_calls[5] == mock.call(
        "[bold yellow]:arrow_forward:[/bold yellow] hello"
    )



"""Support for environment engines."""

from hipercow.environment_engines.base import EnvironmentEngine, Platform
from hipercow.environment_engines.empty import Empty
from hipercow.environment_engines.pip import Pip

__all__ = ["Empty", "EnvironmentEngine", "Pip", "Platform"]



import csv
import platform
import re
import subprocess
from dataclasses import dataclass
from pathlib import Path


# We use 'str' here and not 'Path' because we are interested in
# storing paths that might be of a different type to the host
# operating system (e.g., a windows path while running on a linux
# system, or v.v.). The pathlib package tries to give us posix
# semantics for windows paths and won't let us construct a path of a
# different type than the host system, which means things like leading
# and trailing slashes or forward slashes vs backslashes become hard
# to reason about.
@dataclass
class Mount:
    """The name of the host on which the mount is found."""

    host: str
    """The path of the mount on the host. For windows hosts this will
    be a drive letter, while on unix hosts this will be an absolute
    path (really for both this is an absolute path but convention at
    least means that they're always drives on windows)"""
    remote: str
    """The path on the local machine - i.e., where the network mount
    is found on the machine calling hipercow. This one could actually
    be a Path object and we might move to that.  On windows this is a
    drive letter"""
    local: Path


@dataclass
class PathMap:
    """The mapping between a local path and one on a remote share."""

    path: Path
    """This is the mount that the file can be found on"""
    mount: Mount
    """The location (drive or absolute path) that 'mount' is found on
    the remote machine. This is stored as a 'str' like Mount's
    components, because this must be able to represent a path on a
    different platform to the one currently running hipercow's code"""
    remote: str
    """The path relative to the mount.  We'll make this a str,
    and not a Path, too as this makes testing and reasoning a bit
    easier.  We never include a leading slash but one is implicit on
    windows, and we always store in forward slash form"""
    relative: str


def remap_path(path: Path, mounts: list[Mount]) -> PathMap:
    pos = [m for m in mounts if path.is_relative_to(m.local)]
    if len(pos) > 1:
        msg = "More than one plausible mount for local directory"
        raise Exception(msg)
    elif len(pos) == 0:
        msg = f"Can't map local directory '{path}' to network path"
        raise Exception(msg)
    mount = pos[0]
    relative = path.relative_to(mount.local)
    relative_str = _forward_slash(_drop_leading_slash(str(relative)))

    if m := re.match("^([A-Za-z]:)[/\\\\]?$", str(mount.local)):
        remote = m.group(1).upper()
    elif mount.host in ["qdrive", "wpia-san04"]:
        remote = "Q:"
    else:
        remote = "V:"

    return PathMap(path, mount, remote, relative_str)


def detect_mounts() -> list[Mount]:
    system = platform.system()
    if system == "Windows":
        return _detect_mounts_windows()
    else:
        return _detect_mounts_unix(system)


def _detect_mounts_unix(system: str) -> list[Mount]:
    fstype = _unix_smb_mount_type(system)
    res = subprocess.run(
        ["mount", "-t", fstype], capture_output=True, check=True
    )
    txt = res.stdout.decode("utf-8")
    return [_parse_unix_mount_entry(x) for x in txt.splitlines()]


def _unix_smb_mount_type(platform: str) -> str:
    return "cifs" if platform == "Linux" else "smbfs"


def _parse_unix_mount_entry(x: str) -> Mount:
    pat = re.compile("^//([^@]*@)?([^/]*)/(.*?)\\s+on\\s+(.*?) (.+)$")
    m = pat.match(x)
    if not m:
        msg = f"Failed to match mount output '{x}'"
        raise Exception(msg)

    _, host, remote, local, _ = m.groups()

    return Mount(_clean_dide_hostname(host), remote, Path(local))


def _detect_mounts_windows() -> list[Mount]:
    res = subprocess.run(
        ["powershell", "-c", "Get-SmbMapping|ConvertTo-CSV"],
        capture_output=True,
        check=True,
    )
    txt = res.stdout.decode("utf-8")
    return _parse_windows_mount_output(txt)


def _parse_windows_mount_output(txt: str) -> list[Mount]:
    d = list(csv.reader(txt.splitlines()[1:]))
    header = d[0]
    i_status = header.index("Status")
    i_local = header.index("LocalPath")
    i_remote = header.index("RemotePath")
    return [
        _parse_windows_mount_entry(x[i_local], x[i_remote])
        for x in d[1:]
        if x[i_status] == "OK"
    ]


def _parse_windows_mount_entry(local: str, remote: str) -> Mount:
    m = re.match("^//([^/]+)/(.+)$", _forward_slash(remote))
    if not m:
        msg = "Failed to parse windows entry"
        raise Exception(msg)
    host, remote = m.groups()
    return Mount(_clean_dide_hostname(host), remote, Path(local + "/"))


def _clean_dide_hostname(host: str) -> str:
    return re.sub("\\.dide\\.ic\\.ac\\.uk$", "", host)


def _forward_slash(x: str) -> str:
    return x.replace("\\", "/")


def _backward_slash(x: str) -> str:
    return x.replace("/", "\\")


def _drop_leading_slash(x: str) -> str:
    return re.sub("^[/\\\\]", "", x)



import re

import pytest

from hipercow import root
from hipercow import task_create as tc
from hipercow.configure import configure
from hipercow.resources import TaskResources
from hipercow.task import TaskData, TaskStatus, task_status
from hipercow.util import transient_working_directory


def test_create_simple_task(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)
    assert re.match("^[0-9a-f]{32}$", tid)
    path_data = (
        tmp_path / "hipercow" / "py" / "tasks" / tid[:2] / tid[2:] / "data"
    )
    assert path_data.exists()
    d = TaskData.read(tid, root.open_root(tmp_path))
    assert isinstance(d, TaskData)
    assert d.task_id == tid
    assert d.method == "shell"
    assert d.data == {"cmd": ["echo", "hello world"]}
    assert d.path == "."
    assert d.envvars == {}
    assert d.resources is None


def test_tasks_cannot_be_empty(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with pytest.raises(Exception, match="cannot be empty"):
        with transient_working_directory(tmp_path):
            tc.task_create_shell([], root=r)


def test_submit_with_driver_if_configured(tmp_path, capsys):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    configure("example", root=r)
    capsys.readouterr()
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)
    str1 = capsys.readouterr().out
    assert f"submitting '{tid}'" in str1
    assert task_status(tid, r) == TaskStatus.SUBMITTED


def test_resources_require_configured_driver(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    resources = TaskResources(queue="foo")
    with transient_working_directory(tmp_path):
        with pytest.raises(Exception, match="Can't specify resources"):
            tc.task_create_shell(
                ["echo", "hello world"], resources=resources, root=r
            )


def test_resources_validated_on_submission(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    configure("example", root=r)
    resources = TaskResources(queue="foo")
    with transient_working_directory(tmp_path):
        with pytest.raises(
            Exception, match="Queue 'foo' is not in valid queue list"
        ):
            tc.task_create_shell(
                ["echo", "hello world"], resources=resources, root=r
            )


def test_can_save_resources_on_submission(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    configure("example", root=r)
    resources = TaskResources(memory_per_task=1)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(
            ["echo", "hello world"], resources=resources, root=r
        )
    d = TaskData.read(tid, root.open_root(tmp_path))
    assert d.resources == TaskResources(queue="default", memory_per_task=1)



import pickle
import secrets
import time
from dataclasses import dataclass, field

from hipercow.driver import load_driver
from hipercow.environment import environment_engine
from hipercow.root import OptionalRoot, Root, open_root
from hipercow.util import transient_working_directory


@dataclass
class ProvisioningData:
    name: str
    id: str
    cmd: list[str]
    time: float = field(default_factory=time.time, init=False)

    def write(self, root: Root) -> None:
        path = root.path_provision_data(self.name, self.id)
        path.parent.mkdir(parents=True, exist_ok=False)
        with path.open("wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def read(name: str, id: str, root: Root) -> "ProvisioningData":
        with root.path_provision_data(name, id).open("rb") as f:
            return pickle.load(f)


@dataclass
class ProvisioningResult:
    error: Exception | None
    start: float
    end: float = field(default_factory=time.time, init=False)

    def write(self, name: str, id: str, root: Root) -> None:
        path = root.path_provision_result(name, id)
        with path.open("wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def read(name: str, id: str, root: Root) -> "ProvisioningResult":
        with root.path_provision_result(name, id).open("rb") as f:
            return pickle.load(f)


@dataclass
class ProvisioningRecord:
    data: ProvisioningData
    result: ProvisioningResult | None

    @staticmethod
    def read(name: str, id: str, root: Root) -> "ProvisioningRecord":
        data = ProvisioningData.read(name, id, root)
        try:
            result = ProvisioningResult.read(name, id, root)
        except FileNotFoundError:
            result = None
        return ProvisioningRecord(data, result)


def provision(
    name: str,
    cmd: list[str] | None,
    *,
    driver: str | None = None,
    root: OptionalRoot = None,
) -> None:
    """Provision an environment.

    This function requires that your root has a driver configured
    (with `hipercow.configure`) and an environment created (with
    `hipercow.environment_new`).

    Note that in the commandline tool, this command is grouped into
    the `environment` group; we may move this function into the
    `environment` module in future.

    Args:
        name: The name of the environment to provision

        cmd: Optionally the command to run to do the provisioning. If
            `None` then the environment engine will select an
            appropriate command if it is well defined for your setup.
            The details here depend on the engine.

        driver: The name of the driver to use in provisioning.
            Normally this can be omitted, as `None` (the default) will
            select your driver automatically if only one is
            configured.

        root: The root, or if not given search from the current directory.

    Returns:
        Nothing, called for side effects only.

    """
    root = open_root(root)
    path_config = root.path_environment_config(name)
    if not path_config.exists():
        msg = f"Environment '{name}' does not exist"
        raise Exception(msg)
    dr = load_driver(driver, root)
    # This is a bit gross because we are loading the local platform
    # and not the platform of the target.  We could know that if the
    # driver tells us it (which it could).
    env = environment_engine(name, root)
    id = secrets.token_hex(8)
    with transient_working_directory(root.path):
        cmd = env.check_args(cmd)
    ProvisioningData(name, id, cmd).write(root)
    dr.provision(name, id, root)


def provision_run(name: str, id: str, root: Root) -> None:
    if root.path_provision_result(name, id).exists():
        msg = f"Provisioning task '{id}' for '{name}' has already been run"
        raise Exception(msg)
    data = ProvisioningData.read(name, id, root)
    env = environment_engine(name, root)
    logfile = root.path_provision_log(name, id)
    start = time.time()
    with transient_working_directory(root.path):
        if not env.exists():
            env.create(filename=logfile)
        try:
            env.provision(data.cmd, filename=logfile)
            ProvisioningResult(None, start).write(name, id, root)
        except Exception as e:
            ProvisioningResult(e, start).write(name, id, root)
            msg = "Provisioning failed"
            raise Exception(msg) from e


def provision_history(name: str, root: Root) -> list[ProvisioningRecord]:
    results = [
        ProvisioningRecord.read(name, x.name, root)
        for x in (root.path_environment(name) / "provision").glob("*")
    ]
    results.sort(key=lambda x: x.data.time)
    return results



"""Functions for interacting with tasks."""

import pickle
import re
from dataclasses import dataclass
from enum import Flag, auto

import taskwait

from hipercow.driver import load_driver
from hipercow.resources import TaskResources
from hipercow.root import OptionalRoot, Root, open_root
from hipercow.util import file_create, read_file_if_exists


class TaskStatus(Flag):
    """Status of a task.

    Tasks move from `CREATED` to `SUBMITTED` to `RUNNING` to one of
    `SUCCESS` or `FAILURE`.  In addition a task might be `CANCELLED`
    (this could happen from `CREATED`, `SUBMITTED` or `RUNNING`) or
    might be `MISSING` if it does not exist.

    A runnable task is one that we could use `task_eval` with; it
    might be `CREATED` or `SUBMITTED`.

    A terminal task is one that has reached the latest state it will
    reach, and is `SUCCESS`, `FAILURE` or `CANCELLED`.

    """

    CREATED = auto()
    SUBMITTED = auto()
    RUNNING = auto()
    SUCCESS = auto()
    FAILURE = auto()
    CANCELLED = auto()
    MISSING = auto()
    TERMINAL = SUCCESS | FAILURE | CANCELLED
    RUNNABLE = CREATED | SUBMITTED

    def is_runnable(self) -> bool:
        """Check if a status implies a task can be run."""
        return bool(self & TaskStatus.RUNNABLE)

    def is_terminal(self) -> bool:
        """Check if a status implies a task is completed."""
        return bool(self & TaskStatus.TERMINAL)

    def __str__(self) -> str:
        return str(self.name).lower()


STATUS_FILE_MAP = {
    TaskStatus.SUCCESS: "status-success",
    TaskStatus.FAILURE: "status-failure",
    TaskStatus.CANCELLED: "status-cancelled",
    TaskStatus.RUNNING: "status-running",
    TaskStatus.SUBMITTED: "status-submitted",
}


## TODO: we'll probably move these to use json soon via pydantic.
@dataclass
class TaskTimes:
    created: float
    started: float | None
    finished: float | None

    def write(self, task_id: str, root: Root):
        with root.path_task_times(task_id).open("wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def read(task_id: str, root: Root):
        path_times = root.path_task_times(task_id)
        if path_times.exists():
            with path_times.open("rb") as f:
                return pickle.load(f)
        created = root.path_task_data(task_id).stat().st_ctime
        path_task_running = (
            root.path_task(task_id) / STATUS_FILE_MAP[TaskStatus.RUNNING]
        )
        running = (
            path_task_running.stat().st_ctime
            if path_task_running.exists()
            else None
        )
        return TaskTimes(created, running, None)


def task_exists(task_id: str, root: OptionalRoot = None) -> bool:
    """Test if a task exists.

    A task exists if the `task_id` was used with this hipercow root
    (i.e., if any files associated with it exist).

    Args:
        task_id: The task identifier, a 32-character hex string.
        root: The root, or if not given search from the current directory.

    Returns:
        `True` if the task exists.
    """
    check_task_id(task_id)
    root = open_root(root)
    return root.path_task(task_id).exists()


def task_status(task_id: str, root: OptionalRoot = None) -> TaskStatus:
    """Read task status.

    Args:
        task_id: The task identifier to check, a 32-character hex string.
        root: The root, or if not given search from the current directory.

    Returns:
        The status of the task.
    """
    check_task_id(task_id)
    root = open_root(root)
    path = root.path_task(task_id)
    if not path.exists():
        return TaskStatus.MISSING
    for v, p in STATUS_FILE_MAP.items():
        if (path / p).exists():
            return v
    return TaskStatus.CREATED


def task_log(
    task_id: str, *, outer: bool = False, root: OptionalRoot = None
) -> str | None:
    """Read the task log.

    Not all tasks have logs; tasks that have not yet started (status
    of `CREATED` or `SUBMITTED` and those `CANCELLED` before starting)
    will not have logs, and tasks that were run without capturing
    output will not produce a log either.  Be sure to check if a
    string was returned.

    Args:
        task_id: The task identifier to fetch the log for, a
            32-character hex string.
        outer: Fetch the "outer" logs; these are logs from the
            underlying HPC software before it hands off to hipercow.
        root: The root, or if not given search from the current directory.

    Returns:
        The log as a single string, if present.

    """
    check_task_id(task_id)
    root = open_root(root)
    if not task_exists(task_id, root):
        msg = f"Task '{task_id}' does not exist"
        raise Exception(msg)

    driver = task_driver(task_id, root)
    if not driver:
        if outer:
            msg = "outer logs are only available for tasks with drivers"
            raise Exception(msg)
        return read_file_if_exists(root.path_task_log(task_id))

    dr = load_driver(driver, root)
    return dr.task_log(task_id, outer=outer, root=root)


def set_task_status(
    task_id: str, status: TaskStatus, value: str | None, root: Root
):
    path = root.path_task(task_id) / STATUS_FILE_MAP[status]
    if value is None:
        file_create(path)
    else:
        with path.open("w") as f:
            f.write(value)


@dataclass(kw_only=True)
class TaskData:
    task_id: str
    method: str  # shell etc
    data: dict
    path: str
    environment: str
    resources: TaskResources | None
    envvars: dict[str, str]

    def write(self, root: Root):
        task_data_write(self, root)

    @staticmethod
    def read(task_id: str, root: Root):
        return task_data_read(task_id, root)


def task_data_write(data: TaskData, root: Root) -> None:
    task_id = data.task_id
    path_task_dir = root.path_task(task_id)
    path_task_dir.mkdir(parents=True, exist_ok=True)
    with root.path_task_data(task_id).open("wb") as f:
        pickle.dump(data, f)


def task_data_read(task_id: str, root: Root) -> TaskData:
    with root.path_task_data(task_id).open("rb") as f:
        return pickle.load(f)


@dataclass
class TaskInfo:
    status: TaskStatus
    data: TaskData
    times: TaskTimes


def task_info(task_id: str, root: OptionalRoot = None) -> TaskInfo:
    check_task_id(task_id)
    root = open_root(root)
    status = task_status(task_id, root)
    if status == TaskStatus.MISSING:
        msg = f"Task '{task_id}' does not exist"
        raise Exception(msg)
    data = TaskData.read(task_id, root)
    times = TaskTimes.read(task_id, root)
    return TaskInfo(status, data, times)


def task_list(
    *, root: OptionalRoot = None, with_status: TaskStatus | None = None
) -> list[str]:
    """List known tasks.

    Warning:
      This function could take a long time to execute on large
      projects with many tasks, particularly on large file systems.
      Because the tasks are just returned as a list of strings, it may
      not be terribly useful either.  Think before building a workflow
      around this.

    Args:
      root: The root to search from.
      with_status: Optional status, or set of statuses, to match

    Returns:
      A list of task identifiers.

    """
    root = open_root(root)
    contents = root.path_task(None).rglob("data")
    ids = ["".join(el.parts[-3:-1]) for el in contents if el.is_file()]
    if with_status is not None:
        ids = [i for i in ids if task_status(i, root) & with_status]
    return ids


class TaskWaitWrapper(taskwait.Task):
    def __init__(self, task_id: str, root: Root):
        self.root = root
        self.task_id = task_id
        self.status_waiting = {"created", "submitted"}
        self.status_running = {"running"}

    def status(self) -> str:
        return str(task_status(self.task_id, self.root))

    def log(self) -> list[str] | None:
        value = task_log(self.task_id, root=self.root)
        return value.splitlines() if value else None

    def has_log(self):
        return True


def task_wait(
    task_id: str,
    *,
    root: OptionalRoot = None,
    allow_created: bool = False,
    **kwargs,
) -> bool:
    """Wait for a task to complete.

    Args:
        task_id: The task to wait on.
        root: The root, or if not given search from the current directory.
        allow_created: Allow waiting on a task that has status
            `CREATED`.  Normally this is not allowed because a task
            that is `CREATED` (and not `SUBMITTED`) will not start; if
            you pass `allow_created=True` it is expected that you are
            also manually evaluating this task!
        **kwargs (Any): Additional arguments to `taskwait.taskwait`.

    Returns:
        `True` if the task completes successfully, `False` if it
        fails.  A timeout will throw an error.  We return this boolean
        rather than the `TaskStatus` because this generalises to
        multiple tasks.

    """
    check_task_id(task_id)
    root = open_root(root)
    task = TaskWaitWrapper(task_id, root)

    status = task_status(task_id, root)

    if status == TaskStatus.CREATED and not allow_created:
        msg = f"Cannot wait on task '{task_id}' which has not been submitted"
        raise Exception(msg)

    if status.is_terminal():
        return status == TaskStatus.SUCCESS

    result = taskwait.taskwait(task, **kwargs)
    status = TaskStatus[result.status.upper()]

    return status == TaskStatus.SUCCESS


def task_recent_rebuild(
    *, root: OptionalRoot = None, limit: int | None = None
) -> None:
    """Rebuild the list of recent tasks.

    Args:
        root: The root, or if not given search from the current directory.
        limit: The maximum number of tasks to add to the recent
            list. Use `limit=0` to truncate the list.

    Returns:
        Nothing, called for side effects only.

    """
    root = open_root(root)
    path = root.path_recent()
    if limit is not None and limit == 0:
        if path.exists():
            path.unlink()
        return

    ids = task_list(root=root)
    time = [root.path_task_data(i).stat().st_ctime for i in ids]
    ids = [i for _, i in sorted(zip(time, ids, strict=False))]

    if limit is not None and limit < len(ids):
        ids = ids[-limit:]

    with path.open("w") as f:
        for i in ids:
            f.write(f"{i}\n")


def task_recent(
    *, root: OptionalRoot = None, limit: int | None = None
) -> list[str]:
    """Return a list of recently created tasks.

    Args:
        root: The root, or if not given search from the current directory.
        limit: The maximum number of tasks to return.

    Return:
        A list of task identifiers.  The most recent tasks will be
        **last** in this list (we might change this in a future
        version - yes, that will be annoying).  Note that this is
        recency in **creation**, not **completion**.

    """
    root = open_root(root)
    path = root.path_recent()
    if not path.exists():
        return []

    with path.open() as f:
        ids = [i.strip() for i in f.readlines()]

    id_length = 32
    if not all(len(i) == id_length for i in ids):
        msg = "Recent data list is corrupt, please rebuild"
        raise Exception(msg)

    if limit is not None and limit < len(ids):
        ids = ids[-limit:]

    return ids


def task_last(root: OptionalRoot = None) -> str | None:
    """Return the most recently created task.

    Args:
        root: The root, or if not given search from the current directory.

    Return:
        A task identifier (a 32-character hex string) if any tasks
        have been created (and if the recent task list has not been
        truncated), or `None` if no tasks have been created.

    """
    root = open_root(root)
    task_id = task_recent(limit=1, root=root)
    return task_id[0] if task_id else None


def task_driver(task_id: str, root: Root) -> str | None:
    """Get the driver used to submit a task.

    This may not always be set (e.g., a task was created before a
    driver was configured), in which case we return `None`.

    Args:
        task_id: The task identifier to look up.
        root: The root, or if not given search from the current directory.

    Returns:
        The driver name, if known.  Otherwise `None`.

    """
    path = root.path_task(task_id) / STATUS_FILE_MAP[TaskStatus.SUBMITTED]
    if not path.exists():
        return None
    with path.open() as f:
        value = f.read()
    return value.strip() if value else None


RE_TASK_ID = re.compile("^[0-9a-f]{32}$")


def is_valid_task_id(task_id: str) -> bool:
    return bool(RE_TASK_ID.match(task_id))


def check_task_id(task_id: str) -> None:
    if not is_valid_task_id(task_id):
        msg = f"'{task_id}' does not look like a valid task identifier"
        raise Exception(msg)


def check_task_exists(task_id: str, root: Root):
    if not task_exists(task_id, root):
        msg = f"Task '{task_id}' does not exist"
        raise Exception(msg)



import secrets
import shutil
from pathlib import Path
from string import Template

from taskwait import Task, taskwait

from hipercow import ui
from hipercow.dide.driver import _web_client
from hipercow.dide.mounts import Mount, _backward_slash, detect_mounts
from hipercow.dide.web import DideWebClient
from hipercow.resources import TaskResources
from hipercow.util import read_file_if_exists

BOOTSTRAP = Template(
    r"""@ECHO on
ECHO working directory: %CD%
call set_python_${version2}_64
set PIPX_HOME=\\wpia-hn-app\hipercow\bootstrap-py-windows\python-${version}\pipx
set PIPX_BIN_DIR=\\wpia-hn-app\hipercow\bootstrap-py-windows\python-${version}\bin

echo "Running pipx to install hipercow"
python \\wpia-hn-app\hipercow\bootstrap-py-windows\in\pipx.pyz install ${args} ${target} > \\wpia-hn-app\hipercow\bootstrap-py-windows\in\${bootstrap_id}\${version}.log 2>&1
set ErrorCode=%ERRORLEVEL%
@ECHO ERRORLEVEL was %ErrorCode%
if %ErrorCode% == 0 (
  @ECHO Installation appears to have been successful
) else (
  @ECHO Installation failed
  EXIT /b %ErrorCode%
)
"""  # noqa: E501
)


def bootstrap(
    target: str | None,
    *,
    force: bool = False,
    verbose: bool = False,
    python_versions: list[str] | None = None,
) -> None:
    client = _web_client()
    mount = _bootstrap_mount()

    python_versions = _bootstrap_python_versions(python_versions)
    bootstrap_id = secrets.token_hex(4)
    path = mount.local / _bootstrap_path(bootstrap_id)

    _bootstrap_check_pipx_pyz(path.parent)

    ui.alert_info(f"Bootstrap id: {bootstrap_id}")

    target = _bootstrap_target(target, mount, bootstrap_id)
    args = _bootstrap_args(force=force, verbose=verbose)

    tasks = [
        _bootstrap_submit(client, mount, bootstrap_id, v, target, args)
        for v in python_versions
    ]
    _bootstrap_wait(tasks)
    # We could clean up here with 'shutil.rmtree(path)' but wait until
    # things setle down first, as this removes any hope of debugging,
    # really.


class BootstrapTask(Task):
    def __init__(
        self,
        mount: Mount,
        bootstrap_id: str,
        client: DideWebClient,
        dide_id: str,
        version: str,
    ):
        self.client = client
        self.dide_id = dide_id
        self.version = version
        self.status_waiting = {"created", "submitted"}
        self.status_running = {"running"}
        self.path_log = Path(
            mount.local / _bootstrap_path(bootstrap_id) / f"{version}.log"
        )

    def log(self) -> None:
        pass

    def status(self) -> str:
        return str(self.client.status_job(self.dide_id))

    def has_log(self) -> bool:
        return False


def _bootstrap_submit(
    client: DideWebClient,
    mount: Mount,
    bootstrap_id: str,
    version: str,
    target: str,
    args: str,
) -> BootstrapTask:
    name = f"bootstrap/{bootstrap_id}/{version}"
    path = _bootstrap_path(bootstrap_id) / f"{version}.bat"

    path_local = mount.local / path
    path_local.parent.mkdir(parents=True, exist_ok=True)
    with path_local.open("w") as f:
        f.write(_batch_bootstrap(bootstrap_id, version, target, args))

    resources = TaskResources(queue="AllNodes")  # not BuildQueue, for now
    dide_id = client.submit(_bootstrap_unc(path), name, resources)
    return BootstrapTask(mount, bootstrap_id, client, dide_id, version)


def _bootstrap_target(
    target: str | None, mount: Mount, bootstrap_id: str
) -> str:
    if target is None:
        return "hipercow"
    if not Path(target).exists():
        msg = f"File '{target}' does not exist"
        raise FileNotFoundError(msg)
    dest = _bootstrap_path(bootstrap_id)
    dest_local = mount.local / dest
    dest_local.mkdir(parents=True, exist_ok=True)
    shutil.copy(target, dest_local)
    return _bootstrap_unc(dest / Path(target).name)


def _bootstrap_args(*, force: bool, verbose: bool):
    args = ["--force" if force else "", "--verbose" if verbose else ""]
    return " ".join(args).strip()


def _bootstrap_mount(mounts: list[Mount] | None = None) -> Mount:
    for m in mounts or detect_mounts():
        if m.host == "wpia-hn.hpc" and m.remote == "hipercow":
            return m
    msg = r"Failed to find '\\wpia-hn.hpc\hipercow' in your mounts"
    raise Exception(msg)


def _bootstrap_wait(tasks: list[BootstrapTask]) -> None:
    ui.alert_info(f"Waiting on {len(tasks)} tasks")
    fail = 0
    for t in tasks:
        res = taskwait(t)
        result_str = f"{t.version}: {res.status}"
        if res.status == "success":
            ui.alert_success(result_str)
        else:
            ui.alert_danger(result_str)
        ui.logs("Logs from pipx:", read_file_if_exists(t.path_log), indent=4)
        if res.status != "success":
            ui.logs(
                f"Additional logs from cluster for task '{t.dide_id}':",
                t.client.log(t.dide_id),
                indent=4,
            )
            fail += 1

    if fail:
        msg = f"{fail}/{len(tasks)} bootstrap tasks failed - see logs above"
        raise Exception(msg)


def _batch_bootstrap(
    bootstrap_id: str, version: str, target: str, args: str
) -> str:
    data = {
        "bootstrap_id": bootstrap_id,
        "version": version,
        "version2": version.replace(".", ""),  # Wes: update the batch filenames
        "args": args,
        "target": target,
    }
    return BOOTSTRAP.substitute(data)


def _bootstrap_unc(path: Path):
    path_str = _backward_slash(str(path))
    return f"\\\\wpia-hn\\hipercow\\{path_str}"


def _bootstrap_path(bootstrap_id: str) -> Path:
    return Path("bootstrap-py-windows") / "in" / bootstrap_id


def _bootstrap_check_pipx_pyz(path: Path) -> None:
    if not (path / "pipx.pyz").exists():
        url = "https://github.com/pypa/pipx/releases"
        msg = (
            f"Expected 'pipx.pyz' to be found at '{path}'; download from {url}"
        )
        raise Exception(msg)


def _bootstrap_python_versions(versions: list[str] | None) -> list[str]:
    if not versions:
        # NOTE: duplicates list in hipercow/util.py, we'll tidy this up
        # later too.
        versions = ["3.10", "3.11", "3.12", "3.13"]
    return versions



import pytest

from hipercow import root
from hipercow.configure import (
    _write_configuration,
    configure,
    show_configuration,
    unconfigure,
)
from hipercow.driver import list_drivers, load_driver, load_driver_optional
from hipercow.example import ExampleDriver
from hipercow.task import TaskStatus, task_log, task_status
from hipercow.task_create import task_create_shell
from hipercow.task_eval import task_eval
from hipercow.util import transient_working_directory


def test_no_drivers_are_available_by_default(tmp_path):
    path = tmp_path / "ex"
    root.init(path)
    r = root.open_root(path)
    assert list_drivers(r) == []
    assert load_driver_optional(None, r) is None
    with pytest.raises(Exception, match="No driver configured"):
        load_driver(None, r)
    with pytest.raises(Exception, match="No such driver 'example'"):
        load_driver("example", r)


def test_can_configure_driver(tmp_path):
    path = tmp_path / "ex"
    root.init(path)
    r = root.open_root(path)
    configure("example", root=r)
    assert list_drivers(r) == ["example"]
    assert isinstance(load_driver(None, r), ExampleDriver)


def test_can_unconfigure_driver(tmp_path):
    path = tmp_path / "ex"
    root.init(path)
    r = root.open_root(path)
    configure("example", root=r)
    assert list_drivers(r) == ["example"]
    unconfigure("example", r)
    assert list_drivers(r) == []
    unconfigure("example", r)
    assert list_drivers(r) == []


def test_throw_if_unknown_driver(tmp_path):
    path = tmp_path / "ex"
    root.init(path)
    r = root.open_root(path)
    with pytest.raises(Exception, match="No such driver 'other'"):
        configure("other", root=r)


def test_can_reconfigure_driver(tmp_path, capsys):
    path = tmp_path / "ex"
    root.init(path)
    r = root.open_root(path)
    capsys.readouterr()
    configure("example", root=r)
    str1 = capsys.readouterr().out
    assert "Configured hipercow to use 'example'" in str1
    configure("example", root=r)
    str2 = capsys.readouterr().out
    assert "Updated configuration for 'example'" in str2


def test_get_default_driver(tmp_path):
    path = tmp_path / "ex"
    root.init(path)
    r = root.open_root(path)
    a = ExampleDriver(r)
    a.name = "a"
    b = ExampleDriver(r)
    b.name = "b"
    _write_configuration(a, r)
    _write_configuration(b, r)
    with pytest.raises(Exception, match="More than one candidate driver"):
        load_driver(None, r)


def test_can_show_configuration(tmp_path, capsys):
    path = tmp_path / "ex"
    root.init(path)
    r = root.open_root(path)
    configure("example", root=r)
    capsys.readouterr()
    capsys.readouterr()
    show_configuration(None, r)
    out = capsys.readouterr().out
    assert "Configuration for 'example'" in out
    assert "(no configuration)" in out


def test_can_read_logs_with_example_driver(tmp_path):
    path = tmp_path / "ex"
    root.init(path)
    r = root.open_root(path)
    configure("example", root=r)
    with transient_working_directory(path):
        tid = task_create_shell(["echo", "hello world"], root=r)
    assert task_status(tid, root=r) == TaskStatus.SUBMITTED
    dr = load_driver("example", r)
    assert dr.task_log(tid, root=r) is None
    assert dr.task_log(tid, outer=True, root=r) is None
    with transient_working_directory(path):
        task_eval(tid, capture=True, root=r)
    assert dr.task_log(tid, root=r) == "hello world\n"
    assert dr.task_log(tid, outer=True, root=r) is None
    assert task_log(tid, root=r) == "hello world\n"
    assert task_log(tid, outer=True, root=r) is None


def test_that_example_driver_has_reasonable_resources(tmp_path):
    dr = ExampleDriver(root.init(tmp_path))
    resources = dr.resources()
    assert resources.queues.valid == {"default"}
    assert resources.max_cores == 1
    assert resources.max_memory == 32



import pickle
import time
from dataclasses import dataclass

from hipercow.environment import environment_engine
from hipercow.root import OptionalRoot, Root, open_root
from hipercow.task import (
    TaskData,
    TaskStatus,
    TaskTimes,
    set_task_status,
    task_status,
)


@dataclass
class TaskResult:
    task_id: str
    success: bool
    data: object


def task_eval(
    task_id: str, *, capture: bool, root: OptionalRoot = None
) -> None:
    root = open_root(root)
    data = TaskData.read(task_id, root)
    task_eval_data(data, capture=capture, root=root)


def task_eval_data(data: TaskData, *, capture: bool, root: Root) -> None:
    task_id = data.task_id
    status = task_status(task_id, root)
    if not status.is_runnable():
        msg = f"Can't run '{task_id}', which has status '{status}'"
        raise Exception(msg)

    t_created = root.path_task_data(task_id).stat().st_ctime
    t_start = time.time()

    set_task_status(task_id, TaskStatus.RUNNING, None, root)

    assert data.method == "shell"  # noqa: S101
    res = task_eval_shell(data, capture=capture, root=root)

    t_end = time.time()

    status = TaskStatus.SUCCESS if res.success else TaskStatus.FAILURE
    with root.path_task_result(task_id).open("wb") as f:
        pickle.dump(res.data, f)

    times = TaskTimes(t_created, t_start, t_end)
    times.write(task_id, root)

    set_task_status(task_id, status, None, root)


def task_eval_shell(data: TaskData, *, capture: bool, root: Root) -> TaskResult:
    cmd = data.data["cmd"]
    env = data.envvars
    path = root.path / data.path
    filename = root.path_task_log(data.task_id) if capture else None
    res = environment_engine(data.environment, root).run(
        cmd, check=False, env=env, cwd=path, filename=filename
    )
    success = res.returncode == 0
    return TaskResult(data.task_id, success, None)



import os
from unittest import mock

import pytest

from hipercow import root
from hipercow.configure import configure
from hipercow.environment import environment_new
from hipercow.environment_engines import Pip
from hipercow.provision import provision, provision_history, provision_run
from hipercow.util import file_create


def test_provision_with_example_driver(tmp_path, mocker):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with (r.path / "requirements.txt").open("w") as f:
        f.write("cowsay\n")
    configure("example", root=r)
    mock_run = mock.MagicMock()
    mocker.patch("subprocess.run", mock_run)
    environment_new("default", "pip", r)
    provision("default", [], root=r)

    pr = Pip(r, "default")
    venv_path = str(pr.path())
    env = pr._envvars()

    assert mock_run.call_count == 2
    assert mock_run.mock_calls[0] == mock.call(
        ["python", "-m", "venv", venv_path],
        env=os.environ,
        check=True,
        stderr=mock.ANY,
        stdout=mock.ANY,
    )
    assert mock_run.mock_calls[1] == mock.call(
        ["pip", "install", "--verbose", "-r", "requirements.txt"],
        env=os.environ | env,
        check=True,
        stderr=mock.ANY,
        stdout=mock.ANY,
    )

    h = provision_history("default", r)
    assert len(h) == 1
    assert h[0].result.error is None

    id = h[0].data.id
    with pytest.raises(Exception, match="has already been run"):
        provision_run("default", id, r)

    r.path_provision_result("default", id).unlink()
    h2 = provision_history("default", r)
    assert len(h) == 1
    assert h2[0].result is None


def test_dont_create_on_second_provision(tmp_path, mocker):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    file_create(r.path / "pyproject.toml")
    configure("example", root=r)
    mock_run = mock.MagicMock()
    mocker.patch("subprocess.run", mock_run)

    environment_new("default", "pip", r)
    pr = Pip(r, "default")
    pr.path().mkdir(parents=True)

    provision("default", [], root=r)
    assert mock_run.call_count == 1

    assert mock_run.mock_calls[0] == mock.call(
        ["pip", "install", "--verbose", "."],
        env=os.environ | pr._envvars(),
        check=True,
        stderr=mock.ANY,
        stdout=mock.ANY,
    )


def test_record_provisioning_error(tmp_path, mocker):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    file_create(r.path / "pyproject.toml")
    configure("example", root=r)
    mock_run = mock.MagicMock(side_effect=Exception("some ghastly error"))
    mocker.patch("subprocess.run", mock_run)

    environment_new("default", "pip", r)
    pr = Pip(r, "default")
    pr.path().mkdir(parents=True)

    with pytest.raises(Exception, match="Provisioning failed"):
        provision("default", [], root=r)
    assert mock_run.call_count == 1

    assert mock_run.mock_calls[0] == mock.call(
        ["pip", "install", "--verbose", "."],
        env=os.environ | pr._envvars(),
        check=True,
        stderr=mock.ANY,
        stdout=mock.ANY,
    )

    h = provision_history("default", r)
    assert len(h) == 1
    assert isinstance(h[0].result.error, Exception)


def test_throw_on_provision_if_no_environment(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    configure("example", root=r)
    with pytest.raises(Exception, match="Environment 'default' does not exist"):
        provision("default", [], root=r)



import time
from unittest import mock

import pytest

from hipercow import root
from hipercow import task_create as tc
from hipercow.task import (
    TaskData,
    TaskStatus,
    TaskTimes,
    TaskWaitWrapper,
    check_task_id,
    is_valid_task_id,
    set_task_status,
    task_driver,
    task_exists,
    task_info,
    task_last,
    task_list,
    task_log,
    task_recent,
    task_recent_rebuild,
    task_status,
    task_wait,
)
from hipercow.task_eval import task_eval
from hipercow.util import transient_working_directory


def test_can_check_if_tasks_are_runnable():
    assert TaskStatus.CREATED.is_runnable()
    assert not TaskStatus.CREATED.is_terminal()

    assert not TaskStatus.RUNNING.is_runnable()
    assert not TaskStatus.RUNNING.is_terminal()

    assert not TaskStatus.SUCCESS.is_runnable()
    assert TaskStatus.SUCCESS.is_terminal()


def test_can_set_task_status(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)
    assert task_exists(tid, r)
    assert task_status(tid, r) == TaskStatus.CREATED
    set_task_status(tid, TaskStatus.RUNNING, None, r)
    assert task_status(tid, r) == TaskStatus.RUNNING
    set_task_status(tid, TaskStatus.SUCCESS, None, r)
    assert task_status(tid, r) == TaskStatus.SUCCESS


def test_that_missing_tasks_have_missing_status(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    assert task_status("a" * 32, r) == TaskStatus.MISSING
    assert not task_exists("a" * 32, r)


def test_that_missing_tasks_error_on_log_read(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    task_id = "a" * 32
    with pytest.raises(Exception, match="Task '.+' does not exist"):
        task_log(task_id, root=r)


def test_can_convert_to_nice_string():
    assert str(TaskStatus.CREATED) == "created"


def test_read_task_info(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)
    info = task_info(tid, r)
    assert info.status == TaskStatus.CREATED
    assert info.data == TaskData.read(tid, r)
    assert info.times == TaskTimes.read(tid, r)
    assert isinstance(info.times.created, float)
    assert info.times.started is None
    assert info.times.finished is None


def test_that_missing_tasks_error_on_task_info(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    task_id = "a" * 32
    with pytest.raises(Exception, match="Task '.+' does not exist"):
        task_info(task_id, r)


def test_that_can_read_info_for_completed_task(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)
        task_eval(tid, capture=False, root=r)
        info = task_info(tid, r)
    assert info.status == TaskStatus.SUCCESS
    assert info.data == TaskData.read(tid, r)
    assert info.times == TaskTimes.read(tid, r)
    assert isinstance(info.times.created, float)
    assert isinstance(info.times.started, float)
    assert isinstance(info.times.finished, float)


def test_can_list_tasks(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    assert task_list(root=r) == []
    with transient_working_directory(tmp_path):
        t1 = tc.task_create_shell(["echo", "hello world"], root=r)
    assert task_list(root=r) == [t1]
    with transient_working_directory(tmp_path):
        t2 = tc.task_create_shell(["echo", "hello world"], root=r)
    assert set(task_list(root=r)) == {t1, t2}


def test_can_list_tasks_by_status(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        ids = [tc.task_create_shell(["true"], root=r) for _ in range(5)]
    # 0 is CREATED
    set_task_status(ids[1], TaskStatus.RUNNING, None, r)
    set_task_status(ids[2], TaskStatus.SUCCESS, None, r)
    set_task_status(ids[3], TaskStatus.SUCCESS, None, r)
    set_task_status(ids[4], TaskStatus.FAILURE, None, r)
    assert set(task_list(root=r)) == set(ids)
    assert set(task_list(root=r, with_status=TaskStatus.SUCCESS)) == {
        ids[2],
        ids[3],
    }
    assert set(task_list(root=r, with_status=TaskStatus.TERMINAL)) == {
        ids[2],
        ids[3],
        ids[4],
    }
    assert set(
        task_list(root=r, with_status=TaskStatus.SUCCESS | TaskStatus.RUNNING)
    ) == {ids[1], ids[2], ids[3]}


def test_can_wait_on_completed_task(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)
        task_eval(tid, capture=False, root=r)
    assert task_wait(tid, root=r)


def test_refuse_to_wait_for_created_task(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)
    with pytest.raises(Exception, match="Cannot wait .+ not been submitted"):
        task_wait(tid, root=r)


def test_wait_wrapper_can_get_status(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)
        wrapper = TaskWaitWrapper(tid, r)
        assert wrapper.status() == "created"
        set_task_status(tid, TaskStatus.SUCCESS, None, r)
        assert wrapper.status() == "success"


def test_wait_wrapper_can_get_log(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)
        wrapper = TaskWaitWrapper(tid, r)
        assert wrapper.log() is None
        assert wrapper.has_log()
        task_eval(tid, capture=True, root=r)
        assert wrapper.status() == "success"
        assert wrapper.log() == ["hello world"]
        assert wrapper.has_log()


def test_can_pass_to_task_wait(tmp_path, mocker):
    mock_status = mock.MagicMock(
        side_effect=[TaskStatus.SUBMITTED, TaskStatus.SUCCESS]
    )
    mocker.patch("hipercow.task.task_status", mock_status)
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)
        assert task_wait(tid, root=r)


def test_can_get_last_task_when_none_are_created(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    assert task_last(r) is None
    assert task_recent(root=r) == []


# This might not work wonderfully on windows, because the timing there
# tends to be too coarse.
def test_can_get_recent_tasks(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        ids = [
            tc.task_create_shell(["echo", "hello world"], root=r)
            for _ in range(5)
        ]
    assert task_last(r) == ids[4]
    assert task_recent(root=r) == ids
    assert task_recent(root=r, limit=3) == ids[2:]


def test_can_detect_corrupt_recent_file(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        ids = []
        for i in range(5):
            if i > 0:
                time.sleep(0.01)
            ids.append(tc.task_create_shell(["echo", "hello world"], root=r))
    assert task_recent(root=r) == ids
    with r.path_recent().open("w") as f:
        for i in ids[:2] + [ids[2] + ids[3], ids[4]]:
            f.write(f"{i}\n")
    with pytest.raises(Exception, match="Recent data list is corrupt"):
        task_recent(root=r)
    task_recent_rebuild(root=r)
    assert task_recent(root=r) == ids
    task_recent_rebuild(root=r, limit=3)
    assert task_recent(root=r) == ids[2:]
    task_recent_rebuild(root=r, limit=0)
    assert task_recent(root=r) == []
    task_recent_rebuild(root=r, limit=0)
    assert task_recent(root=r) == []


def test_can_read_driver_for_submitted_task(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)
    assert task_exists(tid, r)
    assert task_status(tid, r) == TaskStatus.CREATED
    assert task_driver(tid, r) is None
    set_task_status(tid, TaskStatus.SUBMITTED, None, r)
    assert task_driver(tid, r) is None
    set_task_status(tid, TaskStatus.SUBMITTED, "example", r)
    assert task_driver(tid, r) == "example"
    set_task_status(tid, TaskStatus.SUCCESS, None, r)
    assert task_driver(tid, r) == "example"


def test_no_outer_log_without_submission(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        tid = tc.task_create_shell(["echo", "hello world"], root=r)
    with pytest.raises(Exception, match="outer logs are only available"):
        task_log(tid, outer=True, root=r)


def test_can_validate_task_id():
    assert is_valid_task_id("3852ea7fe8adab595cc5084d29be0bf7")
    assert not is_valid_task_id("3852ea7fe8adab595cc5084d29be0bf")
    assert not is_valid_task_id("3852ea7fe8adab59ri5cc5084d29be0bf7")
    assert not is_valid_task_id(" 3852ea7fe8adab59ri5cc5084d29be0bf7")
    assert check_task_id("3852ea7fe8adab595cc5084d29be0bf7") is None
    with pytest.raises(Exception, match="does not look like a valid task"):
        check_task_id("3852ea7fe8adab595cc5084d29be")



import pytest

from hipercow import root
from hipercow.environment import (
    EnvironmentConfiguration,
    environment_check,
    environment_delete,
    environment_exists,
    environment_list,
    environment_new,
)


def test_create_pip_environment(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    assert environment_list(r) == ["empty"]
    environment_new("default", "pip", r)
    assert environment_list(r) == ["default", "empty"]
    with pytest.raises(Exception, match="'default' already exists"):
        environment_new("default", "pip", r)
    cfg = EnvironmentConfiguration.read("default", r)
    assert cfg.engine == "pip"


def test_environment_selection(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    assert environment_check("empty", r) == "empty"
    assert environment_check(None, r) == "empty"
    with pytest.raises(Exception, match="No such environment 'default'"):
        environment_check("default", r)
    with pytest.raises(Exception, match="No such environment 'other'"):
        environment_check("other", r)

    environment_new("default", "pip", r)
    assert environment_check("empty", r) == "empty"
    assert environment_check(None, r) == "default"
    assert environment_check("default", r) == "default"
    with pytest.raises(Exception, match="No such environment 'other'"):
        environment_check("other", r)

    environment_new("other", "pip", r)
    assert environment_check("empty", r) == "empty"
    assert environment_check(None, r) == "default"
    assert environment_check("default", r) == "default"
    assert environment_check("other", r) == "other"
    with pytest.raises(Exception, match="No such environment 'other2'"):
        environment_check("other2", r)


def test_delete_environment(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)

    environment_new("default", "pip", r)
    environment_new("other", "pip", r)

    assert environment_list(r) == ["default", "empty", "other"]
    assert environment_exists("other", r)

    environment_delete("other", r)
    assert environment_list(r) == ["default", "empty"]
    assert not environment_exists("other", r)

    environment_delete("default", r)
    assert environment_list(r) == ["empty"]
    assert not environment_exists("default", r)

    with pytest.raises(Exception, match="Can't delete the empty environment"):
        environment_delete("empty", r)


def test_can_only_delete_non_empty_default_environment(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with pytest.raises(Exception, match="as it is empty"):
        environment_delete("default", r)


def test_cant_delete_unknown_environment(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with pytest.raises(Exception, match="as it does not exist"):
        environment_delete("other", r)


def test_require_pip_environment_engine(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with pytest.raises(Exception, match="Only the 'pip' and 'empty'"):
        environment_new("default", "conda", r)



import unittest

import keyring
import pytest

from hipercow.dide import auth


def test_strip_leading_dide_from_username():
    assert auth._check_username("DIDE\\bob") == "bob"
    assert auth._check_username("dide\\bob") == "bob"
    assert auth._check_username("bob") == "bob"
    assert auth._check_username("other\\bob") == "other\\bob"


def test_error_for_invalid_usernames():
    with pytest.raises(Exception, match="empty username"):
        auth._check_username("")
    with pytest.raises(Exception, match="Unexpected newline"):
        auth._check_username("foo\n bar")
    with pytest.raises(Exception, match="Unexpected '#'"):
        auth._check_username("username # your username here")
    with pytest.raises(Exception, match="Unexpected ' '"):
        auth._check_username("username here")


def test_can_get_password(mocker):
    mocker.patch("getpass.getpass", return_value="bob")
    assert auth._get_password() == "bob"


def test_can_error_if_given_empty_password(mocker):
    mocker.patch("getpass.getpass", return_value="")
    with pytest.raises(Exception, match="Invalid empty password"):
        assert auth._get_password()


# These would be much nicer with a fixture for temporary credentials
# (e.g., registering a backend for keyring to use within a context
# manager)
def test_authenticate_flow(mocker):
    mocker.patch("hipercow.dide.auth._default_username", return_value="bob1")
    mocker.patch("hipercow.dide.auth._get_username", return_value="bob")
    mocker.patch("hipercow.dide.auth._get_password", return_value="secret")
    mocker.patch("hipercow.dide.web.check_access", return_value="secret")
    mocker.patch("hipercow.dide.auth.check_access", return_value="secret")
    mocker.patch("keyring.set_password")
    auth.authenticate()
    assert auth._default_username.call_count == 1
    assert auth._get_username.call_count == 1
    assert auth._get_username.call_args == unittest.mock.call("bob1")
    assert auth._get_password.call_count == 1
    assert auth.check_access.call_count == 1
    assert auth.check_access.call_args == unittest.mock.call(
        auth.Credentials("bob", "secret")
    )
    assert auth.keyring.set_password.call_count == 2
    assert auth.keyring.set_password.call_args_list[0] == unittest.mock.call(
        "hipercow/dide/username", "", "bob"
    )
    assert auth.keyring.set_password.call_args_list[1] == unittest.mock.call(
        "hipercow/dide/password", "bob", "secret"
    )


def test_check_flow(mocker):
    creds = unittest.mock.Mock()
    mocker.patch("hipercow.dide.auth.fetch_credentials", return_value=creds)
    mocker.patch("hipercow.dide.auth.check_access")
    auth.check()
    assert auth.fetch_credentials.call_count == 1
    assert auth.check_access.call_count == 1
    assert auth.check_access.call_args == unittest.mock.call(creds)


def test_can_fetch_credentials_from_thekey_ring(mocker):
    mocker.patch("keyring.get_password", side_effect=["bob", "secret"])
    assert auth.fetch_credentials() == auth.Credentials("bob", "secret")
    assert auth.keyring.get_password.call_count == 2
    assert auth.keyring.get_password.call_args_list[0] == unittest.mock.call(
        "hipercow/dide/username", ""
    )
    assert auth.keyring.get_password.call_args_list[1] == unittest.mock.call(
        "hipercow/dide/password", "bob"
    )


def test_can_error_if_username_not_found(mocker):
    mocker.patch("keyring.get_password", return_value=None)
    with pytest.raises(Exception, match="Did not find your DIDE"):
        auth.fetch_credentials()


def test_can_error_if_password_not_found(mocker):
    mocker.patch("keyring.get_password", side_effect=["bob", None])
    with pytest.raises(Exception, match="Did not find your DIDE"):
        auth.fetch_credentials()


def test_can_clear_credentials(mocker):
    mocker.patch("keyring.get_password", return_value="bob")
    mocker.patch("hipercow.dide.auth._delete_password_silently")
    auth.clear()
    assert auth.keyring.get_password.call_count == 1
    assert auth._delete_password_silently.call_count == 2
    assert auth._delete_password_silently.call_args_list[
        0
    ] == unittest.mock.call("hipercow/dide/username", "")
    assert auth._delete_password_silently.call_args_list[
        1
    ] == unittest.mock.call("hipercow/dide/password", "bob")


def test_dont_clear_password_if_no_username_found(mocker):
    mocker.patch("keyring.get_password", return_value="")
    mocker.patch("hipercow.dide.auth._delete_password_silently")
    auth.clear()
    assert auth.keyring.get_password.call_count == 1
    assert auth._delete_password_silently.call_count == 0


def test_delete_password_does_not_error_if_not_found(mocker):
    mocker.patch(
        "keyring.delete_password",
        side_effect=keyring.errors.PasswordDeleteError(),
    )
    auth._delete_password_silently("foo", "bar")
    assert auth.keyring.delete_password.call_count == 1


def test_can_get_default_username(mocker):
    mocker.patch("keyring.get_password", side_effect=["", "bob"])
    mocker.patch("getpass.getuser", return_value="alice")
    assert auth._default_username() == "alice"
    assert auth.getpass.getuser.call_count == 1
    assert auth.keyring.get_password.call_count == 1
    assert auth.keyring.get_password.call_args == unittest.mock.call(
        "hipercow/dide/username", ""
    )
    assert auth._default_username() == "bob"
    assert auth.getpass.getuser.call_count == 1
    assert auth.keyring.get_password.call_count == 2


def test_can_read_username_from_input(mocker):
    mocker.patch("hipercow.dide.auth._get_input", side_effect=["", "bob"])
    assert auth._get_username("alice") == "alice"
    assert auth._get_username("alice") == "bob"



import pytest

from hipercow import root
from hipercow.bundle import (
    bundle_create,
    bundle_delete,
    bundle_list,
    bundle_load,
    bundle_status,
    bundle_status_reduce,
)
from hipercow.task import TaskStatus
from hipercow.task_create import _new_task_id, task_create_shell
from hipercow.util import transient_working_directory


def test_can_create_simple_bundle(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        ids = [task_create_shell(["true"], root=r) for _ in range(5)]
    nm = bundle_create(ids, root=r)
    bundle = bundle_load(nm, root=r)
    assert bundle.name == nm
    assert bundle.task_ids == ids
    assert bundle_list(r) == [nm]


def test_throw_if_load_missing_bundle(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with pytest.raises(Exception, match="No such bundle 'foo'"):
        bundle_load("foo", root=r)


def test_can_delete_bundle(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        ids = [task_create_shell(["true"], root=r) for _ in range(5)]
    nm = bundle_create(ids, root=r)
    assert bundle_list(r) == [nm]
    bundle_delete(nm, root=r)
    assert bundle_list(r) == []


def test_throw_if_delete_missing_bundle(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with pytest.raises(Exception, match="Can't delete bundle 'foo', it"):
        bundle_delete("foo", root=r)


def test_get_bundle_status(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        ids = [task_create_shell(["true"], root=r) for _ in range(5)]
    nm = bundle_create(ids, root=r)
    assert bundle_status(nm, root=r) == [TaskStatus.CREATED] * 5
    assert bundle_status_reduce(nm, root=r) == TaskStatus.CREATED


def test_can_overwrite_bundle(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)
    with transient_working_directory(tmp_path):
        ids = [task_create_shell(["true"], root=r) for _ in range(5)]
    nm = bundle_create(ids[:2], root=r)
    assert bundle_load(nm, root=r).task_ids == ids[:2]
    bundle_create(ids[:3], name=nm, overwrite=True, root=r)
    assert bundle_load(nm, root=r).task_ids == ids[:3]
    with pytest.raises(Exception, match="exists and overwrite is False"):
        bundle_create(ids[:4], name=nm, overwrite=False, root=r)
    assert bundle_load(nm, root=r).task_ids == ids[:3]


def test_can_skip_validation(tmp_path):
    root.init(tmp_path)
    r = root.open_root(tmp_path)

    ids = [_new_task_id() for _ in range(5)]
    with pytest.raises(Exception, match="does not exist"):
        bundle_create(ids, root=r)
    nm = bundle_create(ids, validate=False, root=r)
    assert bundle_load(nm, root=r).task_ids == ids



import datetime
import platform
import re
from pathlib import Path
from string import Template

from hipercow.__about__ import __version__ as version
from hipercow.dide.configuration import DideConfiguration
from hipercow.dide.mounts import PathMap, _backward_slash, _forward_slash
from hipercow.root import Root

TASK_RUN = Template(
    r"""@echo off
REM automatically generated
ECHO generated on host: ${hostname}
ECHO generated on date: ${date}
ECHO hipercow(py) version: ${hipercow_version}
ECHO running on: %COMPUTERNAME%

net use I: \\wpia-hn-app\hipercow /y

call setGit.bat

${network_shares_create}

${hipercow_root_drive}
cd ${hipercow_root_path}
ECHO working directory: %CD%

set HIPERCOW_NO_DRIVERS=1
set HIPERCOW_CORES=%CCP_NUMCPUS%
set REDIS_URL=10.0.2.254

ECHO this is a single task

I:\bootstrap-py-windows\python-${python_version}\bin\hipercow task eval --capture ${task_id}

@ECHO off
set ErrorCode=%ERRORLEVEL%

@REM We could use hipercow here, I think
if exist hipercow\py\tasks\${task_id_1}\${task_id_2}\status-success (
  set TaskStatus=0
) else (
  set TaskStatus=1
)

ECHO ERRORLEVEL was %ErrorCode%

ECHO Cleaning up
%SystemDrive%

${network_shares_delete}

net use I: /delete /y

if %ErrorCode% neq 0 (
  ECHO Task failed catastrophically
  EXIT /b %ErrorCode%
)

if %TaskStatus% == 0 (
  ECHO Task completed successfully!
  ECHO Quitting
) else (
  ECHO Task did not complete successfully
  EXIT /b 1
)"""  # noqa: E501
)

PROVISION = Template(
    r"""@echo off
REM automatically generated
ECHO generated on host: ${hostname}
ECHO generated on date: ${date}
ECHO hipercow(py) version: ${hipercow_version}
ECHO running on: %COMPUTERNAME%

net use I: \\wpia-hn-app\hipercow /y

call setGit.bat

${network_shares_create}

${hipercow_root_drive}
cd ${hipercow_root_path}
ECHO working directory: %CD%

ECHO this is a provisioning task

I:\bootstrap-py-windows\python-${python_version}\bin\hipercow environment provision-run ${environment_name} ${provision_id}

@ECHO off
%SystemDrive%
set ErrorCode=%ERRORLEVEL%

ECHO ERRORLEVEL was %ErrorCode%

ECHO Cleaning up
%SystemDrive%

${network_shares_delete}

net use I: /delete /y

set ERRORLEVEL=%ErrorCode%

if %ERRORLEVEL% neq 0 (
  ECHO Error running provisioning
  EXIT /b %ERRORLEVEL%
)

@ECHO Quitting
"""  # noqa: E501
)


# In a future version, we might prefer to use a configuration object,
# perhaps extracted from the root, rather than the actual argument for
# the path mapping, but this iterates towards something that we can
# start running on the cluster itself before everything is ready.
#
# There's two bits here where we faff about with the unc path - we
# needf the relative path and the absolute path to the task directory
# and we build the unc base path twice (once with slash normalisation,
# the other without).
def write_batch_task_run(
    task_id: str, config: DideConfiguration, root: Root
) -> str:
    data = _template_data_task_run(task_id, config)
    path_map = config.path_map
    path = root.path_task(task_id, relative=True) / "task_run.bat"
    unc = _unc_path(path_map, path)
    with (root.path / path).open("w") as f:
        f.write(TASK_RUN.substitute(data))
    return unc


def write_batch_provision(
    name: str, provision_id: str, config: DideConfiguration, root: Root
) -> str:
    path_map = config.path_map
    data = _template_data_provision(name, provision_id, config)
    path = root.path_provision(name, provision_id, relative=True) / "run.bat"
    unc = _unc_path(path_map, path)

    path_abs = root.path / path
    path_abs.parent.mkdir(parents=True, exist_ok=True)
    with (path_abs).open("w") as f:
        f.write(PROVISION.substitute(data))
    return unc


def _template_data_core(config: DideConfiguration) -> dict[str, str]:
    path_map = config.path_map
    host = _clean_host(path_map.mount.host)
    unc_path = _backward_slash(f"//{host}/{path_map.mount.remote}")
    root_drive = path_map.remote
    root_path = _backward_slash("/" + path_map.relative)

    network_shares_create = f"net use {root_drive} {unc_path} /y"
    network_shares_delete = f"net use {root_drive} /delete /y"

    return {
        "hostname": platform.node(),
        "date": str(datetime.datetime.now(tz=datetime.timezone.utc)),
        "python_version": config.python_version,
        "hipercow_version": version,
        "hipercow_root_drive": root_drive,
        "hipercow_root_path": root_path,
        "network_shares_create": network_shares_create,
        "network_shares_delete": network_shares_delete,
    }


def _template_data_task_run(
    task_id, config: DideConfiguration
) -> dict[str, str]:
    return _template_data_core(config) | {
        "task_id": task_id,
        "task_id_1": task_id[:2],
        "task_id_2": task_id[2:],
    }


def _template_data_provision(
    name: str, id: str, config: DideConfiguration
) -> dict[str, str]:
    return _template_data_core(config) | {
        "environment_name": name,
        "provision_id": id,
    }


def _unc_path(path_map: PathMap, path: Path) -> str:
    path_str = _forward_slash(str(path))
    rel = path_map.relative
    rel = "" if rel == "." else rel + "/"
    host = _clean_host(path_map.mount.host)
    ret = f"//{host}/{path_map.mount.remote}/{rel}{path_str}"
    return _backward_slash(ret)


def _clean_host(host: str) -> str:
    return re.sub("\\.hpc$", "-app", host)



import re
import sys
from functools import reduce
from operator import ior

import click
from click_repl import repl  # type: ignore
from prompt_toolkit.history import FileHistory
from rich.console import Console
from typing_extensions import Never  # 3.10 does not have this in typing

from hipercow import root, ui
from hipercow.bundle import (
    bundle_delete,
    bundle_list,
    bundle_load,
    bundle_status,
    bundle_status_reduce,
)
from hipercow.configure import configure, show_configuration, unconfigure
from hipercow.dide import auth as dide_auth
from hipercow.dide.bootstrap import bootstrap as dide_bootstrap
from hipercow.dide.check import dide_check
from hipercow.driver import list_drivers
from hipercow.environment import (
    environment_delete,
    environment_list,
    environment_new,
)
from hipercow.provision import provision, provision_run
from hipercow.resources import TaskResources
from hipercow.task import (
    TaskStatus,
    task_last,
    task_list,
    task_log,
    task_recent,
    task_recent_rebuild,
    task_status,
    task_wait,
)
from hipercow.task_create import task_create_shell
from hipercow.task_create_bulk import (
    BulkDataInput,
    bulk_create_shell,
    bulk_create_shell_commands,
)
from hipercow.task_eval import task_eval
from hipercow.util import loop_while, read_csv_to_dict, tabulate, truthy_envvar

# This is how the 'rich' docs drive things:
console = Console()


# I (Rich) am surprised that there's no really nice way of doing
# something like this out of the box in click.  The idea that we try
# to implement here:
#
# In most cases where an error is thrown we print the immediate
# exception and set the exit code to 1.
#
# Additionally: if the user sets an environment variable, then we show
# the trace; we can use this for getting additional debug information
# out.
#
# Finally, I've added an option here to just not do anything
# interesting with the output and follow click's normal exception
# handling.
def cli_safe():
    try:
        cli()
    except Exception as e:
        _handle_error(e)


def _handle_error(e: Exception) -> Never:
    if truthy_envvar("HIPERCOW_RAW_ERROR"):
        raise e
    else:
        click.echo(f"Error: {e}")
        if truthy_envvar("HIPERCOW_TRACEBACK"):
            console.print_exception(show_locals=True, suppress=[click])
        else:
            click.echo("For more information, run with 'HIPERCOW_TRACEBACK=1'")
        sys.exit(1)


@click.group()
@click.version_option()
def cli():
    """Interact with hipercow."""
    pass  # pragma: no cover


@cli.command()
@click.argument("path", default=".")
def init(path: str):
    """Initialise a new `hipercow` root.

    Create a new `hipercow` root at the path `path`.  This path should
    be the root directory of your project (e.g., the path containing
    `.git`) and we will create a directory `hipercow/` within that
    directory.

    Once initialised, you should configure a driver and environment.

    The default is to initialise at the current directory, like `git`
    would.

    """
    root.init(path)


@cli.command("repl")
@click.pass_context
def cli_repl(ctx):
    """Launch the interactive REPL.

    Running this creates an interactive session where you can send a
    series of commands to `hipercow`, with nice autocompletion.
    Please let us know how you find this.

    Type `:help` for help within the REPL.  To quit, type `:exit` or
    Ctrl-D.

    """
    prompt_kwargs = {"message": "hipercow> ", "history": _repl_history()}
    loop_while(lambda: _repl_call(ctx, prompt_kwargs))


def _repl_history():
    try:
        r = root.open_root()
        return FileHistory(r.path_repl_history())
    except Exception:
        return None


def _repl_call(ctx, prompt_kwargs) -> bool:
    try:
        repl(ctx, prompt_kwargs=prompt_kwargs)
        return False
    except Exception as e:
        click.echo(f"Error: {e}")
        return True


@cli.group()
def driver():
    """Configure drivers."""
    pass  # pragma: no cover


@driver.command("configure")
@click.argument("name")
@click.option("--python-version", help="Python version to use on the cluster")
def cli_driver_configure(name: str, python_version: str | None):
    """Add a driver.

    This command will initialise a driver.  Most likely you will call

        hipercow driver configure dide-windows

    because `dide-windows` is the only real driver at the moment.  If
    you provide the `--python-version` flag you can specify the Python
    version to use, if you want a version that is different to the
    version of Python that you are using locally.

    """
    configure(name, python_version=python_version)


@driver.command("unconfigure")
@click.argument("name")
def cli_driver_unconfigure(name: str):
    """Unconfigure (remove) a driver."""
    unconfigure(name)


@driver.command("show")
@click.argument("name", required=False)
def cli_driver_show(name: str | None):
    """Show configuration for a driver."""
    show_configuration(name)


@driver.command("list")
def cli_driver_list():
    """List configured drivers."""
    drivers = list_drivers(root.open_root())
    if drivers:
        click.echo("\n".join([str(d) for d in drivers]))
    else:
        click.echo("(none)")


@cli.group()
def task():
    """Create and interact with tasks."""
    pass  # pragma: no cover


@task.command("status")
@click.argument("task_id")
def cli_task_status(task_id: str):
    """Get the status of a task.

    The `task_id` will be a 32-character hex string.  We print a
    single status as a result, this might be `created`, `submitted`,
    `running`, `success` or `failure`.  Additional statuses will be
    added in future as we expand the tool.

    """
    click.echo(task_status(task_id))


@task.command("log")
@click.option(
    "--outer", is_flag=True, help="Print the outer logs, from the HPC system"
)
@click.argument("task_id")
def cli_task_log(task_id: str, *, outer=False):
    """Get a task log.

    If the log does not yet exist, we return nothing.

    """
    value = task_log(task_id, outer=outer)
    if value is not None:
        click.echo(value)


@task.command("list")
@click.option("--with-status", type=str, multiple=True)
def cli_task_list(with_status=None):
    """List all tasks.

    This is mostly meant for debugging; the task list is not very
    interesting and it might take a while to find them all.

    """
    with_status = _process_with_status(with_status)
    for task_id in task_list(with_status=with_status):
        click.echo(task_id)


@task.command("last")
def cli_task_last():
    """List the most recently created task."""
    task_id = task_last()
    if task_id is None:
        # we might set exit code to something nonzero here, but this
        # seems slightly hard...
        pass
    else:
        click.echo(task_id)


@task.command("recent")
@click.option(
    "--limit", type=int, default=10, help="The maximum number of tasks to list"
)
@click.option("--rebuild", is_flag=True, help="Rebuild the recent task list")
def cli_task_recent(limit: int, *, rebuild: bool):
    """List recent tasks."""
    if rebuild:
        task_recent_rebuild(limit=limit)
    for i in task_recent(limit=limit):
        click.echo(i)


@task.command("create")
@click.argument("cmd", nargs=-1)
@click.option(
    "--environment", type=str, help="The environment in which to run the task"
)
@click.option("--queue", help="Queue to submit the task to")
@click.option("--wait", is_flag=True, help="Wait for the task to complete")
def cli_task_create(
    cmd: tuple[str], environment: str | None, *, queue: str | None, wait: bool
):
    """Create a task.

    Submits a command line task to the cluster (if you have a driver
    configured).

    The command can be any shell command, though for complex ones we
    expect that quoting might become interesting - let us know how you
    get on.  If your command involves options (beginning with a `-`)
    you will need to use `--` to separate the commands to `hipercow`
    from those to your application.  For example

    ```
    hipercow task create -- cowsay -t hello
    ```

    which passes the `-t` argument through to `cowsay`.  We may remove this
    requirement in a future version.

    If you have multiple environments, you can specify the environment
    to run the task in with `--environment`.  We validate the presence
    of this environment at task submission.

    If you use `--wait` then we effectively call `hipercow task wait`
    on your newly created task.  You can use this to simulate a
    blocking task create-and-run type loop, but be aware you might
    wait for a very long time if the cluster is busy.

    """
    resources = None if queue is None else TaskResources(queue=queue)
    task_id = task_create_shell(
        list(cmd), environment=environment, resources=resources
    )
    click.echo(task_id)
    if wait:
        task_wait(task_id)


@task.command("eval", hidden=True)
@click.argument("task_id")
@click.option("--capture/--no-capture", default=False)
def cli_task_eval(task_id: str, *, capture: bool):
    task_eval(task_id, capture=capture)


@task.command("wait")
@click.argument("task_id")
@click.option(
    "--poll",
    default=1,
    type=float,
    help="Time to wait between checking on task (in seconds)",
)
@click.option(
    "--timeout", type=float, help="Time to wait for task before failing"
)
@click.option(
    "--show-log/--no-show-log",
    default=True,
    help="Stream logs to the console, if available?",
)
@click.option(
    "--progress/--no-progress",
    default=True,
    help="Show a progress spinner while waiting?",
)
def cli_task_wait(
    task_id: str,
    *,
    poll: float,
    timeout: float,
    show_log: bool,
    progress: bool,
):
    """Wait for a task to complete."""
    task_wait(
        task_id,
        poll=poll,
        timeout=timeout,
        show_log=show_log,
        progress=progress,
    )


def _process_with_status(with_status: list[str]):
    if not with_status:
        return None
    return reduce(ior, [TaskStatus[i.upper()] for i in with_status])


@cli.group()
def dide():
    """Commands for interacting with the DIDE cluster."""
    pass  # pragma: no cover


@dide.command("authenticate")
@click.argument("action", default="set")
def cli_dide_authenticate(action: str):
    """Interact with DIDE authentication.

    The action can be

    * `set`: Set your username and password (the default)

    * `check`: Check the stored credentials

    * `clear`: Clear any stored credentials

    """
    if action == "set":
        dide_auth.authenticate()
    elif action == "check":
        dide_auth.check()
    elif action == "clear":
        dide_auth.clear()
    else:
        msg = f"No such action '{action}'; must be one of set/check/clear"
        raise Exception(msg)


@dide.command("check")
def cli_dide_check():
    """Check everything is good to use hipercow on the DIDE cluster."""
    dide_check()


@dide.command("bootstrap", hidden=True)
@click.argument("target", required=False)
@click.option(
    "--force/--no-force",
    default=False,
    help="Force reinstallation; passed through to pip",
)
@click.option(
    "--verbose/--no-verbose",
    default=True,
    help="Verbose output from pip; default is verbose output",
)
@click.option(
    "--python-version",
    multiple=True,
    help="Python version to update. Multiple copies of this flag allowed",
)
def cli_dide_bootstrap(
    target: str, *, force: bool, verbose: bool, python_version: list[str]
):
    r"""Update the bootstrap.

    You will need `--force` much more often than expcted at present,
    because pip won't always reinstall if only the patch version has
    changed.

    This only works if you have write access to
    `\\wpia-hn\hipercow`.  See the administration guide on the
    hipercow website for details:

    https://mrc-ide.github.io/hipercow-py/administration/

    """
    dide_bootstrap(
        target,
        force=force,
        verbose=verbose,
        python_versions=list(python_version),
    )


@cli.group()
def environment():
    """Interact with environments."""
    pass  # pragma: no cover


@environment.command("list")
def cli_environment_list():
    """List environments."""
    envs = environment_list(root.open_root())
    click.echo("\n".join(envs))


@environment.command("delete")
@click.option("--name")
def cli_environment_delete(name: str):
    """Delete an environment."""
    r = root.open_root()
    environment_delete(name, r)


@environment.command("new")
@click.option("--name", default="default", help="Name of the environment")
@click.option("--engine", default="pip", help="Engine to use")
def cli_environment_new(name: str, engine: str):
    """Create a new environment.

    Note that this does not actually install anything; you will need to use

    ```
    hipercow environment provision
    ```

    to do that, after creation.

    """
    r = root.open_root()
    environment_new(name, engine, r)


@environment.command(
    "provision", context_settings={"ignore_unknown_options": True}
)
@click.option(
    "--name", default="default", help="Name of the environment to provision"
)
@click.argument("cmd", nargs=-1, type=click.UNPROCESSED)
def cli_environment_provision(name: str, cmd: tuple[str]):
    """Provision an environment.

    This will launch a cluster task that installs the packages you
    have requested.  You can pass a command to run here, or use the
    defaults if your project has a well known (and well behaved)
    environment description.

    """
    r = root.open_root()
    provision(name, list(cmd), root=r)


@environment.command("provision-run", hidden=True)
@click.argument("name")
@click.argument("id")
def cli_environment_provision_run(name: str, id: str):
    r = root.open_root()
    provision_run(name, id, r)


@cli.group()
def bundle():
    """Interact with bundles."""
    pass  # pragma: no cover


@bundle.command("list")
def cli_bundle_list():
    """List bundles."""
    r = root.open_root()
    for el in bundle_list(r):
        click.echo(el)


@bundle.command("delete")
@click.argument("name")
def cli_bundle_delete(name: str):
    """Delete a bundle.

    Note that this does not delete the tasks in the bundle, just the
    bundle itself. So you will not be able to use `hipercow bundle` to
    manage the ensemble of jobs together, but you will be able to work
    with the tasks by their individual ids, using `hipercow task`.
    """
    r = root.open_root()
    bundle_delete(name, root=r)


@bundle.command("status")
@click.argument("name")
@click.option(
    "--summary",
    type=click.Choice(["none", "group", "single"], case_sensitive=False),
    default="none",
    help="Summarise the statuses",
)
def cli_bundle_status(name: str, summary: str):
    """Get the status of a bundle.

    This can offer three levels of summary; and we might redesign the
    output a bit to make this easier to work with, depending on what
    people actually do with the output.

    Please don't try and parse the output directly, but let us know
    what sort of format you might like it in, as we can easily add
    something like a JSON format output.

    """
    r = root.open_root()
    if summary == "single":
        click.echo(bundle_status_reduce(name, root=r))
    else:
        res = bundle_status(name, root=r)
        if summary == "none":
            task_ids = bundle_load(name, root=r).task_ids
            for task_id, status in zip(task_ids, res, strict=False):
                click.echo(f"{task_id}: {status}")
        else:
            for status_str, n in tabulate([str(el) for el in res]).items():
                # We might format this more nicely so that we
                # align on status?
                click.echo(f"{status_str}: {n}")


# The names are a bit of a mess here, something that largely follows
# hipercow-r:
#
# hipercow task status <id> (for task.task_status)
# hipercow task create <cmd> (for task_create.task_create_shell)
# hipercow bundle status <name> for (bundle.bundle_status)
# hipercow create bulk <cmd> <data> for (task_create_bulk.bulk_create_shell)
#
# Some of the inconsistency is inherited from hipercow, but it also
# reflects that bundles might not be only created by bulk submission.
#
# I wonder if we might resolve this by having a 'create' subcommand,
# so I've put the bulk submission here and we might move the single
# task creation down here later.  Or we can move elsewhere.
@cli.group("create")
def create():
    pass  # pragma: no cover


@create.command("bulk")
@click.argument("cmd", nargs=-1)
@click.option("--data", multiple=True, help="Data to use in the template")
@click.option(
    "--environment", type=str, help="The environment in which to run the task"
)
@click.option("--queue", help="The queue to submit the task to")
@click.option("--name", help="An optional name for the bundle")
@click.option(
    "--preview",
    help="Show preview of tasks that would be created, but don't create any",
    is_flag=True,
)
def cli_create_bulk(
    cmd: tuple[str],
    *,
    preview: bool,
    environment: str | None,
    data: tuple[str],
    queue: str | None,
    name: str | None,
):
    """Bulk create tasks by substituting into a template.

       These created tasks will then belong to a "bundle"
       with a name (either automatically generated, or of your
       choosing), that can be managed using `hipercow bundle`

    The command must contain `@`-prefixed placeholders such as

    ```
    mycommand --output=path/@{subdir} @action
    ```

    which includes the placeholders `subdir` and `action`.

    You can include data to substitute into this template in three
    ways:

    * A single `--data=filename.csv` argument which will read a `csv` file
      of inputs (here it must contain columns `subdir` and `action`)

    * Two arguments `--data` containing:
      - a comma separated set of values (e.g., `--data action=a,b,c`)
      - a range of values (e.g., `--data subdir=0:n` or `--data
        subdir=0..n`); the `:` form is python-like and does not
        include the end of the range, while the `..` is inclusive and
        does include `n`
      - in both cases we will compute the outer product of all
        `--data` arguments and submit all combinations of arguments.

    """
    template_data = _cli_bulk_create_data(data)
    if preview:
        cmds = bulk_create_shell_commands(list(cmd), template_data)
        _cli_bulk_preview_commands(cmds, 3)
    else:
        r = root.open_root()
        resources = None if queue is None else TaskResources(queue=queue)

        # An alternative here would be to have `bulk_create_shell`
        # *not* take template data and accept a list of lists of
        # strings (i.e., the template with substitutions applied) and
        # then we only do the bulk_create_shell_commands() once here
        # in this function and not in the bulk support.  That feels a
        # bit weird - I think we want the substitution to be part of
        # the programmatic API but it could make this more flexible?
        # We can always change our mind here in future, as we expect
        # most usage to be from the cli.
        name = bulk_create_shell(
            list(cmd),
            template_data,
            name=name,
            environment=environment,
            resources=resources,
            root=r,
        )
        click.echo(name)


def _cli_bulk_create_data(data: tuple[str]) -> BulkDataInput:
    if not data:
        msg = "Expected at least one '--data' argument"
        raise Exception(msg)

    if len(data) == 1 and re.search("\\.csv$", data[0], re.IGNORECASE):
        return read_csv_to_dict(data[0])

    return dict(_cli_bulk_parse_data_argument(el) for el in data)


def _cli_bulk_parse_data_argument(x: str) -> tuple[str, list[str]]:
    m = re.match("^([_a-z][_a-z0-9]*)\\s*=\\s*([^=]+)$", x, re.IGNORECASE)
    if not m:
        msg = f"Failed to parse '--data' argument '{x}'"
        raise Exception(msg)
    name, value = m.groups()
    if m_range := re.match(r"^([0-9]+)(:|\.\.)([0-9]+)$", value):
        start, mode, end = m_range.groups()
        values = range(int(start), int(end) + (0 if mode == ":" else 1))
        return name, [str(i) for i in values]
    return name, list(value.split(","))


def _cli_bulk_preview_commands(cmds: list[list[str]], preview: int) -> None:
    n = len(cmds)
    cmds_show = list(enumerate(cmds))
    skip = n - 2 * preview
    if skip > 0:
        cmds_show = cmds_show[:preview] + cmds_show[-preview:]
    ui.alert_info(f"I would create {n} commands:")
    for i, cmd_i in cmds_show:
        cmd_str = " ".join(cmd_i)
        click.echo(f"  {i + 1}: {cmd_str}")
        if skip > 0 and i == preview - 1:
            click.echo(f"   : ... {skip} commands omitted")



from dataclasses import dataclass

from hipercow.dide.auth import check as check_auth
from hipercow.dide.mounts import Mount, PathMap, remap_path
from hipercow.root import Root
from hipercow.util import check_python_version


@dataclass
class DideConfiguration:
    path_map: PathMap
    python_version: str

    def __init__(
        self,
        root: Root,
        *,
        mounts: list[Mount],
        python_version: str | None = None,
        check_credentials: bool = True,
    ):
        if check_credentials:
            check_auth()
        self.path_map = remap_path(root.path, mounts)
        self.python_version = check_python_version(python_version)



import math

import pytest

from hipercow.resources import ClusterResources, Queues, TaskResources


def test_can_construct_simple_queues():
    x = Queues.simple("foo")
    assert x == Queues({"foo"}, "foo", "foo", "foo")


def test_can_construct_real_queues():
    x = Queues({"a", "b", "c", "d"}, build="b", test="c", default="d")
    assert x.valid == {"a", "b", "c", "d"}
    assert x.test == "c"
    assert x.build == "b"
    assert x.default == "d"


def test_throw_if_queues_are_invalid():
    with pytest.raises(ValueError, match="Build queue 'build' is not"):
        Queues({"a", "b"}, build="build", test="a", default="b")


def test_can_check_user_queue():
    x = Queues({"a", "b", "c", "d"}, build="b", test="c", default="d")
    assert x.validate_queue("a") == "a"
    assert x.validate_queue(None) == "d"
    assert x.validate_queue(".default") == "d"
    assert x.validate_queue(".test") == "c"
    with pytest.raises(ValueError, match="Invalid special queue '.foo'"):
        x.validate_queue(".foo")
    with pytest.raises(ValueError, match="'foo' is not in valid queue list"):
        x.validate_queue("foo")


def test_can_create_resources():
    r = TaskResources()
    assert r.queue is None
    assert r.cores == 1
    assert not r.exclusive
    assert r.max_runtime is None
    assert r.memory_per_node is None
    assert r.memory_per_task is None


def test_can_create_non_default_resources():
    r = TaskResources(
        queue="foo",
        cores=16,
        exclusive=True,
        max_runtime=3600,
        memory_per_node=16,
        memory_per_task=8,
    )
    assert r.queue == "foo"
    assert r.cores == 16
    assert r.exclusive
    assert r.max_runtime == 3600
    assert r.memory_per_node == 16
    assert r.memory_per_task == 8


def test_that_cores_is_positive():
    with pytest.raises(ValueError, match="'cores' must be positive"):
        TaskResources(cores=-1)


def test_that_cores_must_be_integer_or_inf():
    assert TaskResources(cores=math.inf).cores == math.inf
    assert TaskResources(cores=100).cores == 100
    with pytest.raises(ValueError, match="'cores' must be an integer"):
        TaskResources(cores=1.5)


def test_can_validate_resources_against_cluster():
    r = ClusterResources(Queues.simple("default"), max_cores=4, max_memory=256)

    res = r.validate_resources(TaskResources(cores=4, memory_per_node=256))
    assert res == TaskResources(queue="default", cores=4, memory_per_node=256)

    with pytest.raises(ValueError, match="Queue 'foo' is not in valid"):
        r.validate_resources(TaskResources(queue="foo"))

    with pytest.raises(ValueError, match="8 is too many cores"):
        r.validate_resources(TaskResources(cores=8))

    with pytest.raises(ValueError, match="512Gb per task is too large"):
        r.validate_resources(TaskResources(memory_per_task=512))

    with pytest.raises(ValueError, match="257Gb per node is too large"):
        r.validate_resources(TaskResources(memory_per_node=257))



import datetime
import json
import math

import pytest
import responses

from hipercow.__about__ import __version__ as hipercow_version
from hipercow.dide import web
from hipercow.resources import TaskResources
from hipercow.task import TaskStatus


def create_client(*, logged_in=True):
    cl = web.DideWebClient(web.Credentials("", ""))
    if logged_in:
        cl._client._has_logged_in = True
    return cl


@responses.activate
def test_list_headnodes():
    ## https://stackoverflow.com/questions/40361308/create-a-functioning-response-object
    listheadnodes = responses.add(
        responses.POST,
        "https://mrcdata.dide.ic.ac.uk/hpc/_listheadnodes.php",
        body="foo\nbar\n",
        status=200,
    )
    cl = create_client()
    res = cl.headnodes()
    assert res == ["foo", "bar"]

    assert listheadnodes.call_count == 1
    req = listheadnodes.calls[0].request
    assert req.headers["Content-Type"] == "application/x-www-form-urlencoded"
    assert req.body == "user="


def test_can_parse_headnodes_responses():
    assert web._client_parse_headnodes("") == []
    assert web._client_parse_headnodes("foo\n") == ["foo"]
    assert web._client_parse_headnodes("foo\nbar\n") == ["foo", "bar"]


@responses.activate
def test_can_get_task_log():
    payload = '<html><head></head><body onload="document.fsub.submit();"><form name="fsub" id="fsub" action="result.php" method="post"><input type="hidden" id="title" name="title" value="Sm9iIDQ4OTg1MSBGYWlsdXJlIFN0YXR1cw=="/><input type="hidden" id="res" name="res" value="bG9nIGNvbnRlbnRzIQ=="/></form></body></html>\n'  # noqa: E501
    getlog = responses.add(
        responses.POST,
        "https://mrcdata.dide.ic.ac.uk/hpc/showjobfail.php",
        body=payload,
        status=200,
    )
    cl = create_client()
    res = cl.log("1234")
    assert res == "log contents!"

    assert getlog.call_count == 1
    req = getlog.calls[0].request
    assert req.body == "cluster=d3BpYS1obg%3D%3D&hpcfunc=showfail&id=1234"


@responses.activate
def test_can_get_status_for_user():
    payload = """493420	hipercow-py-test	Finished	1 core	DIDE\\rfitzjoh	20250129120445	20250129120445	20250129120446	AllNodes
489851		Failed	1 core	DIDE\\rfitzjoh	20250127160545	20250127160545	20250127160545	LinuxNodes
489823		Finished	1 core	DIDE\\rfitzjoh	20250127160453	20250127160453	20250127160454	LinuxNodes
"""  # noqa: E501
    status = responses.add(
        responses.POST,
        "https://mrcdata.dide.ic.ac.uk/hpc/_listalljobs.php",
        body=payload,
        status=200,
    )
    cl = create_client()
    res = cl.status_user()
    assert len(res) == 3
    assert res[0] == web.DideTaskStatus(
        "493420",
        "hipercow-py-test",
        TaskStatus.SUCCESS,
        "1 core",
        "rfitzjoh",
        datetime.datetime(2025, 1, 29, 12, 4, 45, tzinfo=datetime.timezone.utc),
        datetime.datetime(2025, 1, 29, 12, 4, 45, tzinfo=datetime.timezone.utc),
        datetime.datetime(2025, 1, 29, 12, 4, 46, tzinfo=datetime.timezone.utc),
        "AllNodes",
    )

    assert status.call_count == 1
    req = status.calls[0].request
    assert (
        req.body
        == "user=&scheduler=d3BpYS1obg%3D%3D&state=Kg%3D%3D&jobs=LTE%3D"
    )


@responses.activate
def test_can_get_job_status():
    status = responses.add(
        responses.GET,
        "https://mrcdata.dide.ic.ac.uk/hpc/api/v1/get_job_status/",
        body="Failed",
        status=200,
    )
    cl = create_client()
    res = cl.status_job("1234")
    assert res == TaskStatus.FAILURE
    assert status.call_count == 1


@responses.activate
def test_can_get_software_list():
    payload = {
        "software": [
            {"name": "R", "version": "4.2.3", "call": "setr64_4_2_3.bat"},
            {"name": "R", "version": "4.3.1", "call": "setr64_4_3_1.bat"},
            {"name": "python", "version": "3.11", "call": "python311.bat"},
        ],
        "linuxsoftware": [
            {"name": "R", "version": "4.2.1", "module": "r/4.2.1"},
            {"name": "python", "version": "3.12", "module": "python/3.12"},
        ],
    }
    software = responses.add(
        responses.GET,
        "https://mrcdata.dide.ic.ac.uk/hpc/api/v1/cluster_software",
        body=json.dumps(payload),
        status=200,
    )
    cl = create_client(logged_in=False)
    res = cl.software()
    assert res == {
        "linux": {
            "r": {"4.2.1": {"module": "r/4.2.1"}},
            "python": {"3.12": {"module": "python/3.12"}},
        },
        "windows": {
            "r": {
                "4.2.3": {"call": "setr64_4_2_3.bat"},
                "4.3.1": {"call": "setr64_4_3_1.bat"},
            },
            "python": {"3.11": {"call": "python311.bat"}},
        },
    }
    assert software.call_count == 1


@responses.activate
def test_can_submit_task():
    submit = responses.add(
        responses.POST,
        "https://mrcdata.dide.ic.ac.uk/hpc/submit_1.php",
        body="Job has been submitted. ID: 497979.\n",
        status=200,
    )
    cl = create_client()
    resources = TaskResources(queue="AllNodes")
    res = cl.submit("1234", "myname", resources)
    assert res == "497979"
    assert submit.call_count == 1


@responses.activate
def test_can_cancel_task():
    cancel = responses.add(
        responses.POST,
        "https://mrcdata.dide.ic.ac.uk/hpc/cancel.php",
        body="497979\tWRONG_USER.\n",
        status=200,
    )
    cl = create_client()
    res = cl.cancel("497979")
    assert res == {"497979": "WRONG_USER."}
    assert cancel.call_count == 1


def test_can_check_access():
    with pytest.raises(Exception, match="You do not have access to any"):
        web._client_check_access("wpia-hn", [])
    with pytest.raises(Exception, match="You do not have access to 'wpia-hn'"):
        web._client_check_access("wpia-hn", ["other"])
    with pytest.raises(Exception, match="try one of a, b"):
        web._client_check_access("wpia-hn", ["a", "b"])
    assert web._client_check_access("a", ["a"]) is None
    assert web._client_check_access("a", ["a", "b"]) is None


def test_throw_if_parse_on_submit_fails():
    with pytest.raises(Exception, match="Job submission has failed"):
        web._client_parse_submit("")


def test_wrap_ids_as_list_for_cancel():
    base = {"cluster": web.encode64("cl"), "hpcfunc": web.encode64("cancel")}
    assert web._client_body_cancel("1", "cl") == {"c1": "1", **base}
    assert web._client_body_cancel(["1"], "cl") == {"c1": "1", **base}
    assert web._client_body_cancel(["1", "2"], "cl") == {
        "c1": "1",
        "c2": "2",
        **base,
    }


def test_can_check_if_we_are_logged_in_from_web_client():
    cl = create_client()
    assert cl.logged_in()
    cl = create_client(logged_in=False)
    assert not cl.logged_in()


@responses.activate
def test_can_check_if_we_have_access():
    responses.add(
        responses.POST,
        "https://mrcdata.dide.ic.ac.uk/hpc/_listheadnodes.php",
        body="wpia-hn\n",
        status=200,
    )
    cl = create_client()
    assert cl.check_access() is None


@responses.activate
def test_can_log_in_and_out():
    login = responses.add(
        responses.POST,
        "https://mrcdata.dide.ic.ac.uk/hpc/index.php",
        body="",
        status=200,
    )
    logout = responses.add(
        responses.GET,
        "https://mrcdata.dide.ic.ac.uk/hpc/logout.php",
        status=200,
    )
    cl = create_client(logged_in=False)
    cl.login()
    assert login.call_count == 1
    assert cl.logged_in()
    cl.logout()
    assert login.call_count == 1
    assert logout.call_count == 1
    assert not cl.logged_in()


@responses.activate
def test_throw_if_user_has_no_access():
    body = "<html>You don't seem to have any HPC access</html>"
    responses.add(
        responses.POST,
        "https://mrcdata.dide.ic.ac.uk/hpc/index.php",
        body=body,
        status=200,
    )
    cl = create_client(logged_in=False)
    with pytest.raises(Exception, match="You do not have HPC access"):
        cl.login()


@responses.activate
def test_login_if_using_authenticated_endpoints():
    login = responses.add(
        responses.POST,
        "https://mrcdata.dide.ic.ac.uk/hpc/index.php",
        body="",
        status=200,
    )
    responses.add(
        responses.POST,
        "https://mrcdata.dide.ic.ac.uk/hpc/_listheadnodes.php",
        body="foo\nbar\n",
        status=200,
    )
    cl = create_client(logged_in=False)
    assert cl.headnodes() == ["foo", "bar"]
    assert login.call_count == 1
    assert cl.logged_in()


def test_raise_if_no_access(mocker):
    mocker.patch("hipercow.dide.web.DideWebClient.__init__", return_value=None)
    mocker.patch(
        "hipercow.dide.web.DideWebClient.check_access",
        side_effect=Exception("some error"),
    )
    with pytest.raises(Exception, match="login failed"):
        web.check_access(web.Credentials("a", "b"))


def test_no_error_if_access_is_ok(mocker):
    mocker.patch("hipercow.dide.web.DideWebClient.__init__", return_value=None)
    mocker.patch("hipercow.dide.web.DideWebClient.check_access")
    web.check_access(web.Credentials("a", "b"))


def test_create_basic_submit_data():
    path = r"\\server\share\script.bat"
    resources = TaskResources(queue="AllNodes")
    data = web._client_body_submit(
        path, "job", "windows", resources=resources, workdir=None
    )
    assert data == {
        "cluster": web.encode64("windows"),
        "template": web.encode64("AllNodes"),
        "jn": web.encode64("job"),
        "wd": web.encode64(""),
        "se": web.encode64(""),
        "so": web.encode64(""),
        "jobs": web.encode64(web._call_quote_batch_path(path)),
        "dep": web.encode64(""),
        "hpcfunc": "submit",
        "ver": web.encode64(f"hipercow-py/{hipercow_version}"),
        "rc": web.encode64("1"),
        "rt": web.encode64("Cores"),
    }


def test_can_set_template():
    path = r"\\server\share\script.bat"
    r = TaskResources(queue="BuildQueue")
    r_cmp = TaskResources(queue="AllNodes")
    data = web._client_body_submit(
        path, "job", "windows", resources=r, workdir=None
    )
    data_cmp = web._client_body_submit(
        path, "job", "windows", resources=r_cmp, workdir=None
    )
    assert data == data_cmp | {"template": web.encode64("BuildQueue")}


def test_can_set_resources():
    def build(**kwargs):
        path = r"\\server\share\script.bat"
        resources = TaskResources(queue="AllNodes", **kwargs)
        return web._client_body_submit(
            path, "job", "windows", resources=resources, workdir=None
        )

    default = build()
    assert build(cores=2) == default | {"rc": web.encode64("2")}
    assert build(cores=math.inf) == default | {
        "rc": web.encode64("1"),
        "rt": web.encode64("Nodes"),
    }
    assert build(exclusive=True) == default | {"exc": web.encode64("1")}
    assert build(max_runtime=1000) == default | {"rnt": web.encode64("1000")}
    assert build(memory_per_node=1) == default | {"mpn": web.encode64("1000")}
    assert build(memory_per_task=2) == default | {"epm": web.encode64("2000")}


def test_can_set_workdir():
    path = r"\\server\share\script.bat"
    workdir = r"r\\server\share\some\path"
    resources = TaskResources(queue="AllNodes")
    data = web._client_body_submit(
        path, "job", "windows", resources=resources, workdir=workdir
    )
    data_cmp = web._client_body_submit(
        path, "job", "windows", resources=resources, workdir=None
    )
    assert data == data_cmp | {"wd": web.encode64(workdir)}



"""Support for environments.

This module provides the support needed to implement real environment
engines.  The functions and classes documented here are really for
internal use, but are needed for developing new environment types.  It
may be useful to consult the environment docs on the hipercow
documentation site alongside the documentation here.

"""

import platform
import subprocess
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path

from hipercow.root import Root


# We don't want to call actual 'platform' code very often because our
# intent is that we're generating things for another system.
@dataclass
class Platform:
    """Information about a platform.

    The most basic information about a platform that we need to set up
    an environment, derived from the the `platform` module.

    Attributes:
        system: The name of the system, in lowercase.  Values will be
            `linux`, `windows` or `darwin` (macOS).  We may replace
            this with an `Enum` in future.

        version: The python version, as a 3-element version string.

    """

    system: str
    version: str

    @staticmethod
    def local() -> "Platform":
        """Platform information for the running Python.

        A convenience function to construct suitable platform
        information for the currently running system.

        """
        return Platform(platform.system().lower(), platform.python_version())


class EnvironmentEngine(ABC):
    """Base class for environment engines.

    Attributes:
        root: The `hipercow` root
        name: The name of the environment to provision
        platform: Optionally, the platform to target.

    """

    def __init__(self, root: Root, name: str, platform: Platform | None = None):
        self.root = root
        self.name = name
        self.platform = platform or Platform.local()

    def exists(self) -> bool:
        """Test if an environment exists.

        This method is not abstract, and generally should not need to
        be replaced by derived classes.

        Returns:
            `True` if the environment directory exists, otherwise
            `False`.  Note that `True` does not mean that the
            environment is *usable*; this is mostly intended to be
            used to determine if `create()` needs to be called.

        """
        return self.path().exists()

    def path(self) -> Path:
        """Compute path to the environment contents.

        This base method version will return a suitable path within
        the root.  Implementations can use this path directly (say, if
        the environment path does not need to differ according to
        platform etc), or compute their own.  We might change the
        logic here in future to make this base-class returned path
        more generally useful.

        Returns:
            The path to the directory that will store the environment.

        """
        return self.root.path_environment_contents(self.name)

    @abstractmethod
    def create(self, **kwargs) -> None:
        """Create (or initialise) the environment.

        This method will be called on the target platform, not on the
        client platform.  Most environment systems have a concept of
        initialisation; this will typically create the directory
        referred to by `path()`, and do any required bootstrapping.
        It will not typically install anything for the user.

        In general, we expect `create()` to be called only once per
        environment lifetime, while `provision()` we expect to be
        called every time the environment is modified (one or many
        times).

        Args:
            **kwargs (Any): Additional keyword arguments passed on to
                the concrete method

        Returns:
            Nothing: Called for side-effects only.

        """
        pass  # pragma: no cover

    @abstractmethod
    def check_args(self, cmd: list[str] | None) -> list[str]:
        """Check arguments provided in `cmd` for suitability.

        This method runs on the client; the python process initiating
        the provisioning request, and does not run in the context of
        the process that will create the environment.  In particular,
        don't assume that the platform information is the same.

        Args:
            cmd: A list of arguments to provision an environment, or
                `None` if the user provided none.  In the latter case you
                must provide suitable defaults or error.

        Returns:
            A validated list of arguments.

        """
        pass  # pragma: no cover

    @abstractmethod
    def provision(self, cmd: list[str], **kwargs) -> None:
        """Provision an environment.

        Install packages or software into the environment.

        Args:
            cmd: A command to run in the environment.  Most of the
                time this just calls `hipercow.utils.subprocess_run`
                directly

            **kwargs (Any): Additional keyword arguments passed
                through to the concrete method.

        Returns:
            Nothing: Called for side-effects only.

        """
        pass  # pragma: no cover

    @abstractmethod
    def run(
        self,
        cmd: list[str],
        *,
        env: dict[str, str] | None = None,
        **kwargs,
    ) -> subprocess.CompletedProcess:
        """Run a command within an environment.

        Both provisioning and running tasks will run in their context
        of an environment.  This method must be specialised to
        activate the environment and then run the given shell command.

        This method should (eventually) call
        `hipercow.util.subprocess_run`, returning the value from that
        function.

        Args:
            cmd: The command to run

            env: An optional dictionary of environment variables that
                will be set within the environment.

            **kwargs (Any): Additional methods passed from the
                provisioner or the task runner.

        Return:
            Information about the completed process.  Note that
            errors are not thrown unless the keyword argument
            `check=True` is provided.

        """
        pass  # pragma: no cover
